from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.utils import FORMAT_INSTRUCTION, extract_evaluation_result
from draive.multimodal import Multimodal, MultimodalContent
from draive.stages import Stage

__all__ = ("relevance_evaluator",)


INSTRUCTION: str = f"""\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate the EVALUATED content using solely a relevance metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is relevance - selection of important parts from the REFERENCE content.
The EVALUATED content should include only important information from the REFERENCE avoiding redundancies and excess information.
</EVALUATION_CRITERIA>
{{guidelines}}
<RATING>
Assign a relevance score using exact name of one of the following values:
- "poor" is very low relevance, the content contains mostly irrelevant or redundant information.
- "fair" is low relevance, the content includes some important points but has significant irrelevant parts.
- "good" is moderate relevance, the content covers most important points but includes some unnecessary information.
- "excellent" is high relevance, the content focuses on important information with minor inclusions of less relevant content.
- "perfect" is very high relevance, the content precisely captures only the most important information from the reference.
Use the "none" value for content that cannot be rated at all.
</RATING>

{FORMAT_INSTRUCTION}
"""  # noqa: E501


@evaluator(name="relevance")
async def relevance_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Input was empty!"},
        )

    if not reference:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Reference was empty!"},
        )

    return extract_evaluation_result(
        await Stage.completion(
            MultimodalContent.of(
                "<REFERENCE>",
                reference,
                "</REFERENCE>\n<EVALUATED>",
                evaluated,
                "</EVALUATED>",
            ),
            instructions=INSTRUCTION.format(
                guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n" if guidelines else "",
            ),
        ).execute()
    )



from draive.multimodal.artifact import ArtifactContent
from draive.multimodal.content import (
    Multimodal,
    MultimodalContent,
    MultimodalContentPart,
    MultimodalTag,
)
from draive.multimodal.templates import (
    Template,
    TemplateDeclaration,
    TemplateDefining,
    TemplateListing,
    TemplateLoading,
    TemplateMissing,
    TemplatesRepository,
)
from draive.multimodal.text import TextContent

__all__ = (
    "ArtifactContent",
    "Multimodal",
    "MultimodalContent",
    "MultimodalContentPart",
    "MultimodalTag",
    "Template",
    "TemplateDeclaration",
    "TemplateDefining",
    "TemplateListing",
    "TemplateLoading",
    "TemplateMissing",
    "TemplatesRepository",
    "TextContent",
)



import json
from collections.abc import Generator, Mapping, Sequence
from datetime import UTC, datetime, timedelta
from typing import Any, cast
from uuid import UUID

from haiway import BasicValue, Map, Meta, MetaValues, ctx
from haiway.postgres import Postgres, PostgresRow

from draive.models import (
    ModelContextElement,
    ModelInput,
    ModelMemory,
    ModelMemoryRecall,
    ModelOutput,
)

__all__ = ("PostgresModelMemory",)


def PostgresModelMemory(
    identifier: UUID,
    *,
    recall_limit: int = 0,
    meta: Meta | MetaValues | None = None,
) -> ModelMemory:
    """Create a model memory bound to a Postgres-backed storage.

    Parameters
    ----------
    identifier: UUID
        Key identifying the memory records grouping in the ``memories`` tables.
    recall_limit: int
        Optional maximum number of context elements returned during recall.

    Returns
    -------
    ModelMemory
        Memory interface persisting variables and context elements in Postgres.

    Notes
    ------
    Example schema:
    ```
    CREATE TABLE memories (
        identifier UUID NOT NULL,
        created TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP,
        PRIMARY KEY (identifier)
    );

    CREATE TABLE memories_variables (
        memories UUID NOT NULL REFERENCES memories (identifier) ON DELETE CASCADE,
        variables JSONB NOT NULL DEFAULT '{}'::jsonb,
        created TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP
    );

    CREATE INDEX IF NOT EXISTS
        memories_variables_idx

    ON
        memories_variables (memories, created DESC);

    CREATE TABLE memories_elements (
        identifier UUID NOT NULL DEFAULT gen_random_uuid(),
        memories UUID NOT NULL REFERENCES memories (identifier) ON DELETE CASCADE,
        content JSONB NOT NULL,
        created TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP,
        PRIMARY KEY (identifier)
    );

    CREATE INDEX IF NOT EXISTS
        memories_elements_idx

    ON
        memories_elements (
            memories,
            created DESC
        );
    ```
    """
    assert recall_limit >= 0  # nosec: B101

    async def recall(
        **extra: Any,
    ) -> ModelMemoryRecall:
        ctx.log_info(f"Recalling memory for {identifier}...")

        ctx.log_info("...loading variables...")
        variables: Mapping[str, BasicValue] = await _load_variables(
            identifier=identifier,
        )

        ctx.log_info("...loading context...")

        context_elements: Sequence[ModelContextElement] = await _load_context(
            identifier=identifier,
            limit=recall_limit,
        )

        ctx.log_info(f"...{len(context_elements)} context elements recalled!")
        ctx.record_info(
            event="postgres.memory.recall",
            attributes={
                "identifier": str(identifier),
                "limit": recall_limit,
                "elements": len(context_elements),
                "variables": len(variables),
            },
        )

        return ModelMemoryRecall(
            context=context_elements,
            variables=variables,
        )

    async def remember(
        *elements: ModelContextElement,
        variables: Mapping[str, BasicValue] | None = None,
        **extra: Any,
    ) -> None:
        if not elements and variables is None:
            return ctx.log_info(f"No content to remember for {identifier}, skipping!")

        ctx.log_info(f"Remembering content for {identifier}...")

        async with Postgres.acquire_connection() as connection:
            async with connection.transaction():
                if variables is not None:
                    ctx.log_info(f"...remembering {len(variables)} variables...")
                    await connection.execute(
                        """
                        INSERT INTO
                            memories_variables (
                                memories,
                                variables
                            )

                        VALUES (
                            $1::UUID,
                            $2::JSONB
                        );
                        """,
                        identifier,
                        json.dumps(variables),
                    )

                ctx.log_info(f"...remembering {len(elements)} context elements...")
                timestamp: datetime = datetime.now(UTC)
                for idx, element in enumerate(elements):
                    created: datetime = timestamp + timedelta(microseconds=idx)

                    await connection.execute(
                        """
                        INSERT INTO
                            memories_elements (
                                identifier,
                                memories,
                                content,
                                created
                            )

                        VALUES (
                            $1::UUID,
                            $2::UUID,
                            $3::JSONB,
                            $4::TIMESTAMPTZ
                        )

                        ON CONFLICT (identifier) DO UPDATE
                            SET
                                content = EXCLUDED.content

                            WHERE
                                memories_elements.memories = EXCLUDED.memories;
                        """,
                        element.identifier,
                        identifier,
                        element.to_json(),
                        created,
                    )

        ctx.record_info(
            event="postgres.memory.remember",
            attributes={
                "identifier": str(identifier),
                "elements": len(elements),
                "variables": len(variables) if variables is not None else 0,
            },
        )
        ctx.log_info("...memory persisted!")

    async def maintenance(
        variables: Mapping[str, BasicValue] | None = None,
        **extra: Any,
    ) -> None:
        ctx.log_info(f"Performing maintenance for {identifier}...")
        async with Postgres.acquire_connection() as connection:
            async with connection.transaction():
                ctx.log_info("...ensuring memory entry exists...")
                await connection.execute(
                    """
                        INSERT INTO
                            memories (
                                identifier
                            )

                        VALUES (
                            $1::UUID
                        )

                        ON CONFLICT (identifier)
                        DO NOTHING;
                        """,
                    identifier,
                )

                if variables is not None:
                    ctx.log_info(f"...remembering {len(variables)} variables...")
                    await connection.execute(
                        """
                        INSERT INTO
                            memories_variables (
                                memories,
                                variables
                            )

                        VALUES (
                            $1::UUID,
                            $2::JSONB
                        );
                        """,
                        identifier,
                        json.dumps(variables),
                    )

        ctx.log_info("...maintenance completed!")

    return ModelMemory(
        recalling=recall,
        remembering=remember,
        maintaining=maintenance,
        meta=Meta.of(meta if meta is not None else {"source": "postgres"}),
    )


async def _load_context(
    *,
    identifier: UUID,
    limit: int,
) -> Sequence[ModelContextElement]:
    rows: Sequence[PostgresRow]
    if limit:
        rows = await Postgres.fetch(
            """
            SELECT
                content::JSONB

            FROM
                memories_elements

            WHERE
                memories = $1::UUID

            ORDER BY
                created DESC

            LIMIT $2;
            """,
            identifier,
            limit,
        )

    else:
        rows = await Postgres.fetch(
            """
            SELECT
                content::JSONB

            FROM
                memories_elements

            WHERE
                memories = $1::UUID

            ORDER BY
                created DESC;
            """,
            identifier,
        )

    return tuple(_decode_context(rows))


def _decode_context(
    rows: Sequence[PostgresRow],
    /,
) -> Generator[ModelContextElement]:
    trim_prefix: bool = True
    skip_output: bool = True
    for row in reversed(rows):
        decoded: Mapping[str, BasicValue] = json.loads(cast(str, row["content"]))
        match decoded.get("type"):
            case "model_input":
                model_input: ModelInput = ModelInput.from_mapping(decoded)
                if trim_prefix:
                    if model_input.contains_tools:
                        skip_output = True
                        continue

                    else:
                        skip_output = False
                        trim_prefix = False

                yield model_input

            case "model_output":
                model_output: ModelOutput = ModelOutput.from_mapping(decoded)
                if trim_prefix:
                    if model_output.contains_tools:
                        continue

                    elif skip_output:
                        continue

                    else:
                        trim_prefix = False

                yield model_output

            case other:
                raise ValueError(f"Invalid model context element: {other}")


async def _load_variables(
    *,
    identifier: UUID,
) -> Mapping[str, BasicValue]:
    row: PostgresRow | None = await Postgres.fetch_one(
        """
        SELECT
            variables::JSONB

        FROM
            memories_variables

        WHERE
            memories = $1::UUID

        ORDER BY created DESC

        LIMIT 1;
        """,
        identifier,
    )

    if row is None:
        return Map()

    return cast(Mapping[str, BasicValue], json.loads(cast(str, row["variables"])))



from typing import Any

from draive.models import (
    GenerativeModel,
    ModelInput,
    ModelInstructions,
    ModelOutput,
)
from draive.multimodal import MultimodalContent
from draive.resources import ResourceContent, ResourceReference

__all__ = ("generate_image",)


async def generate_image(
    *,
    instructions: ModelInstructions,
    input: MultimodalContent,  # noqa: A002
    **extra: Any,
) -> ResourceContent | ResourceReference:
    result: ModelOutput = await GenerativeModel.completion(
        instructions=instructions,
        context=[ModelInput.of(input)],
        output="image",
        stream=False,
        **extra,
    )

    for image in result.content.images():
        return image

    raise ValueError("Failed to generate a valid image")



import re
from collections.abc import Callable, Generator, Mapping, MutableSequence
from typing import Final

from draive.multimodal.content import Multimodal, MultimodalContent

# Match `{%variable%}` where the name contains no whitespace.
_PLACEHOLDER_PATTERN: Final[re.Pattern[str]] = re.compile(r"{%([^\s%]+)%}")

__all__ = (
    "parse_template_variables",
    "resolve_multimodal_template",
    "resolve_text_template",
)


def parse_template_variables(
    template: str,
) -> Generator[str]:
    for match in _PLACEHOLDER_PATTERN.finditer(template):
        yield match.group(1)


def resolve_text_template(
    template: str,
    *,
    arguments: Mapping[str, Multimodal],
) -> str:
    parts: MutableSequence[str] = []
    append: Callable[[str], None] = parts.append
    get_argument: Callable[[str], Multimodal | None] = arguments.get
    cursor: int = 0

    for match in _PLACEHOLDER_PATTERN.finditer(template):
        start: int
        end: int
        start, end = match.span()
        if start > cursor:
            append(template[cursor:start])

        variable_name: str = match.group(1)
        variable_value: Multimodal | None = get_argument(variable_name)
        if variable_value is None:
            raise KeyError(f"Missing template argument: {variable_name}")

        if isinstance(variable_value, str):
            append(variable_value)

        else:
            append(variable_value.to_str())

        cursor = end

    if cursor < len(template):
        append(template[cursor:])

    return "".join(parts)


def resolve_multimodal_template(
    template: str,
    *,
    arguments: Mapping[str, Multimodal],
) -> MultimodalContent:
    parts: MutableSequence[Multimodal] = []
    append: Callable[[Multimodal], None] = parts.append
    get_argument: Callable[[str], Multimodal | None] = arguments.get
    cursor: int = 0

    for match in _PLACEHOLDER_PATTERN.finditer(template):
        start, end = match.span()
        if start > cursor:
            append(template[cursor:start])

        variable_name: str = match.group(1)
        variable_value: Multimodal | None = get_argument(variable_name)
        if variable_value is None:
            raise KeyError(f"Missing template argument: {variable_name}")

        append(variable_value)
        cursor = end

    if cursor < len(template):
        append(template[cursor:])

    if not parts:
        return MultimodalContent.empty

    return MultimodalContent.of(*parts)



from typing import Any, Protocol, Self, final, overload, runtime_checkable

from haiway import State, ctx, statemethod

__all__ = (
    "Memory",
    "MemoryMaintaining",
    "MemoryRecalling",
    "MemoryRemembering",
)


@runtime_checkable
class MemoryRecalling[Recalled](Protocol):
    async def __call__(
        self,
        **extra: Any,
    ) -> Recalled: ...


@runtime_checkable
class MemoryRemembering[Remembered](Protocol):
    async def __call__(
        self,
        *items: Remembered,
        **extra: Any,
    ) -> None: ...


@runtime_checkable
class MemoryMaintaining(Protocol):
    async def __call__(
        self,
        **extra: Any,
    ) -> None: ...


async def _remembering_none(
    *items: Any,
    **extra: Any,
) -> None:
    pass  # noop


async def _maintaining_noop(
    **extra: Any,
) -> None:
    pass  # noop


@final
class Memory[Recalled, Remembered](State):
    @classmethod
    def constant(
        cls,
        recalled: Recalled,
    ) -> Self:
        async def recalling(
            **extra: Any,
        ) -> Recalled:
            return recalled

        return cls(
            recalling=recalling,
            remembering=_remembering_none,
        )

    @overload
    @classmethod
    async def recall(
        cls,
        **extra: Any,
    ) -> Recalled: ...

    @overload
    async def recall(
        self,
        **extra: Any,
    ) -> Recalled: ...

    @statemethod
    async def recall(
        self,
        **extra: Any,
    ) -> Recalled:
        ctx.record_info(
            event="memory.recall",
        )
        return await self.recalling(**extra)

    @overload
    @classmethod
    async def remember(
        cls,
        *items: Remembered,
        **extra: Any,
    ) -> None: ...

    @overload
    async def remember(
        self,
        *items: Remembered,
        **extra: Any,
    ) -> None: ...

    @statemethod
    async def remember(
        self,
        *items: Remembered,
        **extra: Any,
    ) -> None:
        ctx.record_info(
            event="memory.remember",
            attributes={"items": len(items)},
        )
        await self.remembering(*items, **extra)

    @overload
    @classmethod
    async def maintenance(
        cls,
        **extra: Any,
    ) -> None: ...

    @overload
    async def maintenance(
        self,
        **extra: Any,
    ) -> None: ...

    @statemethod
    async def maintenance(
        self,
        **extra: Any,
    ) -> None:
        ctx.record_info(
            event="memory.maintenance",
        )
        await self.maintaining(**extra)

    recalling: MemoryRecalling[Recalled]
    remembering: MemoryRemembering[Remembered]
    maintaining: MemoryMaintaining = _maintaining_noop



from collections.abc import Iterable
from typing import Any, overload

from haiway import State, ctx, statemethod

from draive.generation.text.default import generate_text
from draive.generation.text.types import TextGenerating
from draive.models import ModelInstructions, Tool, Toolbox
from draive.multimodal import Multimodal, MultimodalContent, Template, TemplatesRepository

__all__ = ("TextGeneration",)


class TextGeneration(State):
    @overload
    @classmethod
    async def generate(
        cls,
        *,
        instructions: Template | ModelInstructions = "",
        input: Template | Multimodal,
        tools: Toolbox | Iterable[Tool] = (),
        examples: Iterable[tuple[Multimodal, str]] = (),
        **extra: Any,
    ) -> str: ...

    @overload
    async def generate(
        self,
        *,
        instructions: Template | ModelInstructions = "",
        input: Template | Multimodal,
        tools: Toolbox | Iterable[Tool] = (),
        examples: Iterable[tuple[Multimodal, str]] = (),
        **extra: Any,
    ) -> str: ...

    @statemethod
    async def generate(
        self,
        *,
        instructions: Template | ModelInstructions = "",
        input: Template | Multimodal,  # noqa: A002
        tools: Toolbox | Iterable[Tool] = (),
        examples: Iterable[tuple[Multimodal, str]] = (),
        **extra: Any,
    ) -> str:
        async with ctx.scope("generate_text"):
            if isinstance(instructions, Template):
                ctx.record_info(
                    attributes={"instructions.template": instructions.identifier},
                )

            if isinstance(input, Template):
                ctx.record_info(
                    attributes={"input.template": input.identifier},
                )

            return await self.generating(
                # resolve instructions templates
                instructions=await TemplatesRepository.resolve_str(instructions),
                # resolve input templates
                input=await TemplatesRepository.resolve(input),
                toolbox=Toolbox.of(tools),
                examples=((MultimodalContent.of(ex_in), ex_out) for ex_in, ex_out in examples),
                **extra,
            )

    generating: TextGenerating = generate_text



from collections.abc import AsyncGenerator, AsyncIterable
from datetime import UTC, datetime
from typing import Any, Literal, overload

from haiway import ctx

from draive.conversation.types import ConversationMessage, ConversationOutputChunk
from draive.models import (
    GenerativeModel,
    ModelContextElement,
    ModelInput,
    ModelInstructions,
    ModelMemory,
    ModelMemoryRecall,
    ModelOutput,
    ModelReasoning,
    ModelToolRequest,
    Toolbox,
)
from draive.multimodal import ArtifactContent

__all__ = ("conversation_completion",)


@overload
async def conversation_completion(
    *,
    instructions: ModelInstructions,
    toolbox: Toolbox,
    memory: ModelMemory,
    input: ConversationMessage,
    stream: Literal[False] = False,
    **extra: Any,
) -> ConversationMessage: ...


@overload
async def conversation_completion(
    *,
    instructions: ModelInstructions,
    toolbox: Toolbox,
    memory: ModelMemory,
    input: ConversationMessage,
    stream: Literal[True],
    **extra: Any,
) -> AsyncIterable[ConversationOutputChunk]: ...


async def conversation_completion(
    *,
    instructions: ModelInstructions,
    toolbox: Toolbox,
    memory: ModelMemory,
    input: ConversationMessage,  # noqa: A002
    stream: bool = False,
    **extra: Any,
) -> AsyncIterable[ConversationOutputChunk] | ConversationMessage:
    """Default conversation completion using ``GenerativeModel.loop``.

    Normalizes memory, resolves instructions, and runs a tool-enabled loop over the
    generative model, returning a message or streaming output chunks.
    """
    if stream:
        return ctx.stream(
            _conversation_completion_stream,
            instructions=instructions,
            toolbox=toolbox,
            memory=memory,
            input=input,
            **extra,
        )

    else:
        return await _conversation_completion(
            instructions=instructions,
            toolbox=toolbox,
            memory=memory,
            input=input,
            **extra,
        )


async def _conversation_completion(
    instructions: ModelInstructions,
    toolbox: Toolbox,
    memory: ModelMemory,
    input: ConversationMessage,  # noqa: A002
    **extra: Any,
) -> ConversationMessage:
    # relying on memory recall correctness
    memory_recall: ModelMemoryRecall = await memory.recall()
    context: list[ModelContextElement] = [
        *memory_recall.context,
        ModelInput.of(input.content),
    ]

    # run input moderation in parallel - TODO: should we use sanitized input?
    result: ModelOutput = await GenerativeModel.loop(
        instructions=instructions,
        toolbox=toolbox,
        context=context,
        stream=False,
        **extra,
    )

    ctx.log_debug("...finalizing message...")
    response_message: ConversationMessage = ConversationMessage.model(
        created=datetime.now(UTC),
        content=result.content_with_reasoning,
    )

    ctx.log_debug("...remembering...")
    try:
        await memory.remember(*context[len(memory_recall.context) :])

    except Exception as exc:
        ctx.log_error(
            "Failed to remember conversation context",
            exception=exc,
        )
        raise exc

    ctx.log_debug("... response message finished!")
    return response_message


async def _conversation_completion_stream(
    instructions: ModelInstructions,
    toolbox: Toolbox,
    memory: ModelMemory,
    input: ConversationMessage,  # noqa: A002
    **extra: Any,
) -> AsyncGenerator[ConversationOutputChunk]:
    memory_recall: ModelMemoryRecall = await memory.recall()
    context: list[ModelContextElement] = [
        *memory_recall.context,
        ModelInput.of(input.content),
    ]

    async for chunk in await GenerativeModel.loop(
        instructions=instructions,
        toolbox=toolbox,
        context=context,
        stream=True,
        **extra,
    ):
        if isinstance(chunk, ModelReasoning):
            # Wrap reasoning as an artifact to stream alongside content
            yield ConversationOutputChunk.of(
                ArtifactContent.of(
                    chunk,
                    category="reasoning",
                    hidden=True,
                )
            )

        else:
            assert not isinstance(chunk, ModelToolRequest)  # nosec: B101
            yield ConversationOutputChunk.of(chunk)

    ctx.log_debug("...remembering...")
    try:
        await memory.remember(*context[len(memory_recall.context) :])

    except Exception as exc:
        ctx.log_error(
            "Failed to remember conversation context",
            exception=exc,
        )
        raise exc

    ctx.log_debug("... streaming message finished!")



from typing import Literal

from cohere import AsyncClientV2
from haiway import getenv_str

__all__ = ("CohereAPI",)


class CohereAPI:
    __slots__ = (
        "_api_key",
        "_aws_region",
        "_base_url",
        "_client",
        "_provider",
        "_timeout",
    )

    def __init__(
        self,
        provider: Literal["cohere", "bedrock"] = "cohere",
        /,
        *,
        base_url: str | None = None,
        api_key: str | None = None,
        aws_region: str | None = None,
    ) -> None:
        self._provider: Literal["cohere", "bedrock"] = provider
        self._base_url: str | None = base_url or getenv_str("COHERE_SERVER_URL")
        self._api_key: str | None = api_key or getenv_str("COHERE_API_KEY")
        self._aws_region: str | None = aws_region or getenv_str("AWS_BEDROCK_REGION")

        self._client: AsyncClientV2 = self._prepare_client()

    def _prepare_client(self) -> AsyncClientV2:
        match self._provider:
            case "cohere":
                return AsyncClientV2(
                    api_key=self._api_key,
                    base_url=self._base_url,
                )

            case "bedrock":
                from draive.cohere.bedrock import AsyncBedrockClientV2

                return AsyncBedrockClientV2(aws_region=self._aws_region)

    async def _initialize_client(self) -> None:
        await self._deinitialize_client()

        self._client = self._prepare_client()
        await self._client.__aenter__()  # pyright: ignore[reportUnknownMemberType]

    async def _deinitialize_client(self) -> None:
        await self._client.__aexit__(None, None, None)  # pyright: ignore[reportUnknownMemberType]



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.utils import FORMAT_INSTRUCTION, extract_evaluation_result
from draive.multimodal import Multimodal, MultimodalContent
from draive.stages import Stage

__all__ = ("readability_evaluator",)


INSTRUCTION: str = f"""\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Carefully examine provided CONTENT, then rate it using solely a readability metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is readability - the ease with which a reader can understand the content.
A readable content uses clear and concise language, is well-structured,
and avoids complex or convoluted elements.
</EVALUATION_CRITERIA>
{{guidelines}}
<RATING>
Assign a readability score using exact name of one of the following values:
- "poor" is very low readability, the content is extremely difficult to understand, with complex language and convoluted structure.
- "fair" is low readability, the content is challenging to read, with frequent use of complex sentences, unclear language or irrelevant parts.
- "good" is moderate readability, the content is somewhat clear but has some areas that are difficult to understand.
- "excellent" is high readability, the content is mostly clear and easy to read, with minor instances of complexity.
- "perfect" is very high readability, the content is highly clear, concise, and easy to understand throughout.
Use the "none" value for content that cannot be rated at all.
</RATING>

{FORMAT_INSTRUCTION}
"""  # noqa: E501


@evaluator(name="readability")
async def readability_evaluator(
    evaluated: Multimodal,
    /,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Input was empty!"},
        )

    return extract_evaluation_result(
        await Stage.completion(
            MultimodalContent.of(
                "<CONTENT>",
                evaluated,
                "</CONTENT>",
            ),
            instructions=INSTRUCTION.format(
                guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n" if guidelines else "",
            ),
        ).execute()
    )



from collections.abc import Generator, Iterable, Mapping
from typing import Any, final

from haiway import BasicValue, State, ctx

from draive.conversation.realtime.default import realtime_conversation_preparing
from draive.conversation.realtime.types import (
    RealtimeConversationPreparing,
    RealtimeConversationSessionScope,
)
from draive.conversation.types import ConversationMessage
from draive.models import (
    ModelInput,
    ModelInstructions,
    ModelMemory,
    ModelMemoryRecall,
    ModelOutput,
    ModelSessionOutputSelection,
    Tool,
    Toolbox,
)
from draive.multimodal import Template, TemplatesRepository

__all__ = ("RealtimeConversation",)


@final
class RealtimeConversation(State):
    """Helper for preparing realtime conversation sessions.

    Normalizes tools and memory into model context and delegates to the configured
    realtime preparation implementation.
    """

    @classmethod
    async def prepare(
        cls,
        *,
        instructions: Template | ModelInstructions = "",
        tools: Toolbox | Iterable[Tool] = (),
        memory: ModelMemory | Iterable[ConversationMessage] = (),
        output: ModelSessionOutputSelection = "auto",
        **extra: Any,
    ) -> RealtimeConversationSessionScope:
        """Prepare a realtime conversation session scope.

        Parameters
        ----------
        instructions : Template | ModelInstructions, optional
            Instructions to steer the session.
        tools : Toolbox | Iterable[Tool], optional
            Tools to expose within the session.
        memory : ModelMemory | Iterable[ConversationMessage], optional
            Initial memory; an iterable is converted to model context.
        output : ModelSessionOutputSelection, optional
            Desired session output selection policy.
        **extra : Any
            Provider-specific kwargs forwarded to the preparation implementation.
        """

        async with ctx.scope("conversation_realtime"):
            conversation: RealtimeConversation = ctx.state(cls)

            conversation_memory: ModelMemory
            memory_variables: Mapping[str, BasicValue] | None
            if isinstance(memory, ModelMemory):
                conversation_memory = memory
                memory_recall: ModelMemoryRecall = await memory.recall()
                memory_variables = memory_recall.variables

            else:

                def model_context_elements() -> Generator[ModelInput | ModelOutput]:
                    for message in memory:
                        if message.role == "user":
                            yield ModelInput.of(message.content)

                        else:
                            yield ModelOutput.of(message.content)

                conversation_memory = ModelMemory.constant(*model_context_elements())
                memory_variables = None

            if isinstance(instructions, Template):
                ctx.record_info(
                    attributes={"instructions.template": instructions.identifier},
                )

            return await conversation.preparing(
                instructions=await TemplatesRepository.resolve_str(
                    instructions,
                    arguments={
                        key: value if isinstance(value, str) else str(value)
                        for key, value in memory_variables.items()
                    }
                    if memory_variables
                    else None,
                ),
                toolbox=Toolbox.of(tools),
                memory=conversation_memory,
                output=output,
                **extra,
            )

    preparing: RealtimeConversationPreparing = realtime_conversation_preparing



from draive.guardrails.safety.regex import RegexSanitizationRule, guardrails_regex_sanitizer
from draive.guardrails.safety.state import GuardrailsSafety
from draive.guardrails.safety.types import GuardrailsSafetyException, GuardrailsSafetySanitization

__all__ = (
    "GuardrailsSafety",
    "GuardrailsSafetyException",
    "GuardrailsSafetySanitization",
    "RegexSanitizationRule",
    "guardrails_regex_sanitizer",
)



import asyncio
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Literal, cast, overload
from uuid import uuid4

from haiway import AttributePath, AttributeRequirement, as_dict, as_list
from qdrant_client.models import (
    CollectionsResponse,
    Datatype,
    Distance,
    Filter,
    FilterSelector,
    PayloadSchemaType,
    PointStruct,
    VectorParams,
)

from draive.embedding import Embedded
from draive.parameters import DataModel
from draive.qdrant.filters import prepare_filter
from draive.qdrant.session import QdrantSession
from draive.qdrant.types import QdrantPaginationResult, QdrantPaginationToken

__all__ = ("QdrantStoreMixin",)


class QdrantStoreMixin(QdrantSession):
    async def collections(self) -> Sequence[str]:
        collections: CollectionsResponse = await self.client.get_collections()
        return tuple(collection.name for collection in collections.collections)

    async def create_collection[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        vector_size: int,
        vector_type: Literal["float32", "float16", "uint8"] | None = None,
        distance: Literal[
            "Cosine",
            "Euclid",
            "Dot",
            "Manhattan",
        ] = "Cosine",
        in_ram: bool,
        skip_existing: bool,
        **extra: Any,
    ) -> bool:
        if skip_existing and await self.client.collection_exists(collection_name=model.__name__):
            return False

        return await self.client.create_collection(
            collection_name=model.__name__,
            vectors_config=VectorParams(
                size=vector_size,
                datatype=Datatype(vector_type) if vector_type is not None else None,
                distance=Distance(distance),
                on_disk=not in_ram,
            ),
            on_disk_payload=True,
            **extra,
        )

    async def create_payload_index[Model: DataModel, Attribute](
        self,
        model: type[Model],
        /,
        *,
        path: AttributePath[Model, Attribute] | Attribute,
        index_type: Literal[
            "keyword",
            "integer",
            "float",
            "geo",
            "text",
            "bool",
            "datetime",
            "uuid",
        ],
        **extra: Any,
    ) -> bool:
        if not await self.client.collection_exists(collection_name=model.__name__):
            return False

        await self.client.create_payload_index(
            collection_name=model.__name__,
            field_name=str(path),
            field_type=PayloadSchemaType(index_type),
            wait=True,
            **extra,
        )
        return True

    async def delete_collection[Model: DataModel](
        self,
        model: type[Model],
        /,
    ) -> None:
        await self.client.delete_collection(collection_name=model.__name__)

    @overload
    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        continuation: QdrantPaginationToken | None = None,
        limit: int = 32,
        include_vector: Literal[False] = False,
        **extra: Any,
    ) -> QdrantPaginationResult[Model]: ...

    @overload
    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        continuation: QdrantPaginationToken | None = None,
        limit: int = 32,
        include_vector: Literal[True],
        **extra: Any,
    ) -> QdrantPaginationResult[Embedded[Model]]: ...

    @overload
    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        continuation: QdrantPaginationToken | None = None,
        limit: int = 32,
        include_vector: bool,
        **extra: Any,
    ) -> QdrantPaginationResult[Embedded[Model]] | QdrantPaginationResult[Model]: ...

    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        continuation: QdrantPaginationToken | None = None,
        limit: int = 32,
        include_vector: bool = False,
        **extra: Any,
    ) -> QdrantPaginationResult[Embedded[Model]] | QdrantPaginationResult[Model]:
        records, next_point_id = await self.client.scroll(
            collection_name=model.__name__,
            scroll_filter=prepare_filter(requirements=requirements),
            limit=limit,
            offset=continuation.next_id if continuation else None,
            with_payload=True,
            with_vectors=include_vector,
            **extra,
        )

        continuation_token: QdrantPaginationToken | None
        if next_point_id is not None:
            continuation_token = QdrantPaginationToken(next_id=next_point_id)

        else:
            continuation_token = None

        if include_vector:
            return QdrantPaginationResult[Embedded[model]](
                results=[
                    Embedded[model](
                        value=model.from_mapping(record.payload),
                        # we are using only a single vector
                        vector=cast(list[float], record.vector),
                    )
                    for record in records
                    if record.payload is not None
                ],
                continuation_token=continuation_token,
            )

        else:
            return QdrantPaginationResult[model](
                results=[
                    model.from_mapping(record.payload)
                    for record in records
                    if record.payload is not None
                ],
                continuation_token=continuation_token,
            )

    async def store[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        objects: Iterable[Embedded[Model]],
        batch_size: int = 64,
        max_retries: int = 3,
        parallel_tasks: int = 1,
        **extra: Any,
    ) -> None:
        await asyncio.to_thread(
            self._partial_store(
                model,
                objects=objects,
                batch_size=batch_size,
                max_retries=max_retries,
                parallel_tasks=parallel_tasks,
                **extra,
            ),
        )

    def _partial_store[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        objects: Iterable[Embedded[Model]],
        batch_size: int = 64,
        max_retries: int = 3,
        parallel_tasks: int = 1,
        **extra: Any,
    ) -> Callable[[], None]:
        def store() -> None:
            # upload_points is a bit weird - spawns multiple threads/processes
            # while blocking current one, we need to offload it to executor
            self.client.upload_points(
                collection_name=model.__name__,
                points=[
                    PointStruct(
                        id=uuid4().hex,
                        payload=as_dict(element.value.to_mapping()),
                        vector=as_list(element.vector),
                    )
                    for element in objects
                ],
                batch_size=batch_size,
                max_retries=max_retries,
                parallel=parallel_tasks,
                method="spawn",
                wait=True,
                **extra,
            )

        return store

    async def delete[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None,
        **extra: Any,
    ) -> None:
        await self.client.delete(
            collection_name=model.__name__,
            points_selector=FilterSelector(
                filter=prepare_filter(
                    requirements=requirements,
                    default=Filter(),
                ),
            ),
            wait=True,
            **extra,
        )



from draive.aws.client import AWS
from draive.aws.observability import CloudwatchObservability
from draive.aws.state import AWSSQS, AWSCloudwatch
from draive.aws.types import AWSAccessDenied, AWSError, AWSResourceNotFound

__all__ = (
    "AWS",
    "AWSSQS",
    "AWSAccessDenied",
    "AWSCloudwatch",
    "AWSError",
    "AWSResourceNotFound",
    "CloudwatchObservability",
)



from collections.abc import Sequence
from typing import Any, Literal, overload
from uuid import UUID

from haiway import AttributeRequirement, as_list
from qdrant_client.conversions.common_types import ScoredPoint

from draive.parameters import DataModel
from draive.qdrant.filters import prepare_filter
from draive.qdrant.session import QdrantSession
from draive.qdrant.types import QdrantResult

__all__ = ("QdrantSearchMixin",)


class QdrantSearchMixin(QdrantSession):
    @overload
    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: Literal[False] = False,
        **extra: Any,
    ) -> Sequence[Model]: ...

    @overload
    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: Literal[True],
        **extra: Any,
    ) -> Sequence[QdrantResult[Model]]: ...

    @overload
    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: bool,
        **extra: Any,
    ) -> Sequence[QdrantResult[Model]] | Sequence[Model]: ...

    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: bool = False,
        **extra: Any,
    ) -> Sequence[QdrantResult[Model]] | Sequence[Model]:
        results: list[ScoredPoint] = await self.client.search(
            collection_name=model.__name__,
            query_filter=prepare_filter(requirements=requirements),
            query_vector=as_list(query_vector),
            score_threshold=score_threshold,
            limit=limit,
            with_payload=True,
            with_vectors=include_vector,
            **extra,
        )

        if include_vector:
            return tuple(
                _qdrant_result(
                    model,
                    data=result,
                )
                for result in results
            )

        else:
            return tuple(
                model(**result.payload) for result in results if result.payload is not None
            )


def _qdrant_result[Content: DataModel](
    content: type[Content],
    /,
    data: ScoredPoint,
) -> QdrantResult[Content]:
    if data.payload is None:
        raise ValueError("Missing qdrant data payload")

    identifier: UUID
    if isinstance(data.id, int):
        identifier = UUID(int=data.id)

    else:
        assert isinstance(data.id, str)  # nosec: B101
        identifier = UUID(hex=data.id)

    return QdrantResult(
        identifier=identifier,
        vector=_flat_vector(data.vector),
        score=data.score,
        content=content.from_mapping(data.payload),
    )


def _flat_vector(
    vector: Any,
    /,
) -> Sequence[float]:
    match vector:
        case [*vector]:
            assert all(isinstance(element, float) for element in vector)  # nosec: B101
            return tuple(vector)

        case None:
            raise ValueError("Missing qdrant data vector")

        case _:
            raise ValueError("Unsupported qdrant data vector")



from draive.qdrant.client import QdrantClient
from draive.qdrant.index import QdrantVectorIndex
from draive.qdrant.state import Qdrant
from draive.qdrant.types import (
    QdrantException,
    QdrantPaginationResult,
    QdrantPaginationToken,
    QdrantResult,
)

__all__ = (
    "Qdrant",
    "QdrantClient",
    "QdrantException",
    "QdrantPaginationResult",
    "QdrantPaginationToken",
    "QdrantResult",
    "QdrantVectorIndex",
)



from collections.abc import Iterable
from typing import Any

from draive.models import (
    GenerativeModel,
    ModelInput,
    ModelInstructions,
    ModelOutput,
    Toolbox,
)
from draive.multimodal import MultimodalContent

__all__ = ("generate_text",)


async def generate_text(
    *,
    instructions: ModelInstructions,
    input: MultimodalContent,  # noqa: A002
    toolbox: Toolbox,
    examples: Iterable[tuple[MultimodalContent, str]],
    **extra: Any,
) -> str:
    result: ModelOutput = await GenerativeModel.loop(
        instructions=instructions,
        context=[
            *[
                message
                for example in examples
                for message in [
                    ModelInput.of(example[0]),
                    ModelOutput.of(MultimodalContent.of(example[1])),
                ]
            ],
            ModelInput.of(input),
        ],
        toolbox=toolbox,
        output="text",
        stream=False,
        **extra,
    )

    return result.content.to_str()



import mimetypes
import re
from asyncio import gather
from collections.abc import Collection, Mapping, Sequence
from io import BytesIO
from pathlib import Path
from typing import Any
from urllib.parse import ParseResult, urlparse

from botocore.exceptions import ClientError  # pyright: ignore[reportMissingModuleSource]
from haiway import META_EMPTY, BasicValue, Meta, asynchronous

from draive.aws.api import AWSAPI
from draive.aws.types import AWSAccessDenied, AWSError, AWSResourceNotFound
from draive.resources import ResourceContent, ResourceReference

__all__ = ("AWSS3Mixin",)


class AWSS3Mixin(AWSAPI):
    async def list(
        self,
        *,
        uri: str | None = None,
        recursive: bool = True,
        limit: int | None = None,
        **extra: Any,
    ) -> Sequence[ResourceReference]:
        """List buckets or objects stored under an S3 bucket or prefix.

        Parameters
        ----------
        uri
            Optional S3 URI pointing to the bucket or prefix
            (``s3://bucket/prefix``). When omitted, lists buckets.
        recursive
            When ``True`` (default), list all objects under the prefix.
            When ``False``, list only the immediate children and prefix markers.
        limit
            Optional maximum number of results to return.

        Returns
        -------
        Sequence[ResourceReference]
            References to objects and prefix markers under the requested scope.

        Raises
        ------
        ValueError
            If the ``uri`` does not use the ``s3://`` scheme.
        AWSAccessDenied
            If S3 rejects the request due to missing or invalid credentials.
        AWSResourceNotFound
            If the bucket does not exist.
        AWSError
            For other S3 client failures.
        """
        if uri is None:
            return await self._list_buckets(
                limit=limit,
                **extra,
            )

        if not uri.startswith("s3://"):
            raise ValueError("Unsupported list uri scheme")

        parsed_uri: ParseResult = urlparse(uri)  # s3://bucket/prefix
        bucket: str = parsed_uri.netloc
        prefix: str = parsed_uri.path.lstrip("/")
        return await self._list(
            bucket=bucket,
            prefix=prefix,
            recursive=recursive,
            limit=limit,
            **extra,
        )

    async def fetch(
        self,
        uri: str,
        **extra: Any,
    ) -> Collection[ResourceReference] | ResourceContent | None:
        """Fetch an object from S3 and wrap it in ``ResourceContent``.

        Parameters
        ----------
        uri
            S3 URI pointing to the object (``s3://bucket/key``).

        Returns
        -------
        Collection[ResourceReference] | ResourceContent | None
            Downloaded content wrapped with metadata for S3 objects.

        Raises
        ------
        ValueError
            If the ``uri`` does not use the ``s3://`` scheme.
        AWSAccessDenied
            If S3 rejects the request due to missing or invalid credentials.
        AWSResourceNotFound
            If the bucket or object key does not exist.
        AWSError
            For other S3 client failures.
        """
        if not uri.startswith("s3://"):
            raise ValueError("Unsupported fetch uri scheme")

        parsed_uri: ParseResult = urlparse(uri)  # s3://bucket/name
        bucket: str = parsed_uri.netloc
        name: str = parsed_uri.path.lstrip("/")
        content, object_info = await gather(
            self._fetch(
                bucket=bucket,
                name=name,
            ),
            self._get_object_info(
                bucket=bucket,
                name=name,
            ),
            return_exceptions=False,
        )
        mime_type, meta = object_info
        return ResourceContent.of(
            content,
            mime_type=mime_type or "application/octet-stream",
            meta=meta,
        )

    @asynchronous
    def _fetch(
        self,
        bucket: str,
        name: str,
    ) -> bytes:
        output: BytesIO = BytesIO()
        try:
            self._s3_client.download_fileobj(
                Bucket=bucket,
                Key=name,
                Fileobj=output,
            )

        except ClientError as exc:
            raise _translate_client_error(
                error=exc,
                bucket=bucket,
                key=name,
            ) from exc

        return output.getvalue()

    async def download(
        self,
        uri: str,
        /,
        *,
        save_path: Path | str,
    ) -> None:
        """Download an S3 object directly to disk.

        Parameters
        ----------
        uri
            S3 URI pointing to the object (``s3://bucket/key``).
        save_path
            Filesystem path where the object should be written.

        Raises
        ------
        ValueError
            If the ``uri`` does not use the ``s3://`` scheme.
        AWSAccessDenied
            If S3 rejects the request due to missing or invalid credentials.
        AWSResourceNotFound
            If the bucket or object key does not exist.
        AWSError
            For other S3 client failures.
        """
        if not uri.startswith("s3://"):
            raise ValueError("Unsupported download uri scheme")

        parsed_uri: ParseResult = urlparse(uri)  # s3://bucket/name
        bucket: str = parsed_uri.netloc
        name: str = parsed_uri.path.lstrip("/")
        return await self._download(
            bucket=bucket,
            name=name,
            save_path=save_path,
        )

    @asynchronous
    def _download(
        self,
        bucket: str,
        name: str,
        save_path: Path | str,
    ) -> None:
        with open(save_path, "wb") as file:
            try:
                self._s3_client.download_fileobj(
                    Bucket=bucket,
                    Key=name,
                    Fileobj=file,
                )

            except ClientError as exc:
                raise _translate_client_error(
                    error=exc,
                    bucket=bucket,
                    key=name,
                ) from exc

    @asynchronous
    def _get_object_info(
        self,
        bucket: str,
        name: str,
    ) -> tuple[str | None, Meta]:
        try:
            response: Any = self._s3_client.head_object(
                Bucket=bucket,
                Key=name,
            )

        except ClientError as exc:
            raise _translate_client_error(
                error=exc,
                bucket=bucket,
                key=name,
            ) from exc

        else:
            return response.get("ContentType"), Meta.of(response.get("Metadata"))

    async def upload(
        self,
        uri: str,
        content: ResourceContent,
        **extra: Any,
    ) -> Meta:
        """Upload ``ResourceContent`` to an S3 location.

        Parameters
        ----------
        uri
            Destination S3 URI (``s3://bucket/key``).
        content
            Data and metadata to upload.

        Returns
        -------
        Meta
            Empty metadata placeholder; retained for interface parity
            with repository interactions.

        Raises
        ------
        ValueError
            If the ``uri`` does not use the ``s3://`` scheme.
        AWSAccessDenied
            If S3 rejects the request due to missing or invalid credentials.
        AWSResourceNotFound
            If the bucket does not exist.
        AWSError
            For other S3 client failures.
        """
        if not uri.startswith("s3://"):
            raise ValueError("Unsupported upload uri scheme")

        parsed_uri: ParseResult = urlparse(uri)  # s3://bucket/name
        bucket: str = parsed_uri.netloc
        name: str = parsed_uri.path.lstrip("/")
        await self._upload(
            bucket=bucket,
            name=name,
            content=content.to_bytes(),
            content_type=content.mime_type,
            meta=content.meta,
        )

        return META_EMPTY

    @asynchronous
    def _list(  # noqa: C901, PLR0912
        self,
        *,
        bucket: str,
        prefix: str,
        recursive: bool,
        limit: int | None,
        **extra: Any,
    ) -> Sequence[ResourceReference]:
        try:
            paginator = self._s3_client.get_paginator("list_objects_v2")
            pagination_config: dict[str, int] = {}
            if limit is not None:
                pagination_config["MaxItems"] = limit
                pagination_config["PageSize"] = min(1000, limit)

            request: dict[str, Any] = {
                "Bucket": bucket,
                "Prefix": prefix,
            }
            if not recursive:
                request["Delimiter"] = "/"

            if pagination_config:
                request["PaginationConfig"] = pagination_config

            if extra:
                request.update(extra)

            references: list[ResourceReference] = []
            for page in paginator.paginate(**request):
                for entry in page.get("Contents", []):
                    if limit is not None and len(references) >= limit:
                        break

                    key = entry.get("Key", "")
                    if not key or (key == prefix and key.endswith("/")):
                        continue

                    references.append(
                        _object_reference(
                            bucket=bucket,
                            key=key,
                            entry=entry,
                        )
                    )

                if limit is not None and len(references) >= limit:
                    break

                for common_prefix in page.get("CommonPrefixes", []):
                    if limit is not None and len(references) >= limit:
                        break

                    prefix_key = common_prefix.get("Prefix", "")
                    if not prefix_key:
                        continue

                    references.append(
                        ResourceReference.of(
                            f"s3://{bucket}/{prefix_key}",
                            name=_object_name(prefix_key),
                            meta=Meta.of({"type": "prefix"}),
                        )
                    )

                if limit is not None and len(references) >= limit:
                    break

            return references

        except ClientError as exc:
            raise _translate_client_error(
                error=exc,
                bucket=bucket,
                key=prefix,
            ) from exc

    @asynchronous
    def _list_buckets(
        self,
        *,
        limit: int | None,
        **extra: Any,
    ) -> Sequence[ResourceReference]:
        try:
            response: Mapping[str, Any] = self._s3_client.list_buckets(**extra)
            buckets: Collection[Mapping[str, Any]] = response.get("Buckets", ())
            references: list[ResourceReference] = [
                _bucket_reference(bucket) for bucket in buckets if bucket.get("Name")
            ]
            if limit is not None:
                return references[:limit]

            return references

        except ClientError as exc:
            error_info: Mapping[str, Any] = getattr(exc, "response", {}).get("Error", {})
            code: str = str(error_info.get("Code") or "").strip()
            message: str = str(error_info.get("Message") or str(exc)).strip()
            raise AWSError(
                uri="s3://",
                code=code or None,
                message=message,
            ) from exc

    @asynchronous
    def _upload(
        self,
        bucket: str,
        name: str,
        content: bytes,
        content_type: str | None = None,
        meta: Mapping[str, BasicValue] | None = None,
    ) -> None:
        try:
            self._s3_client.put_object(
                Bucket=bucket,
                Body=content,
                Key=name,
                ContentType=content_type if content_type else None,
                Metadata=_sanitize_metadata(meta),
            )

        except ClientError as exc:
            raise _translate_client_error(
                error=exc,
                bucket=bucket,
                key=name,
            ) from exc


def _sanitize_metadata_value(value: Any) -> str:
    # Convert to string first
    text: str = str(value)

    # Replace newlines and tabs with spaces
    text = re.sub(r"[\n\r\t]+", " ", text)

    # Remove any control characters
    text = re.sub(r"[\x00-\x1F\x7F-\x9F]", "", text)

    # Collapse multiple spaces
    text = re.sub(r"\s+", " ", text).strip()

    # S3 metadata value limit is 1,024 bytes per value
    max_bytes = 1021  # Reserve 3 bytes for "..."
    text_bytes = text.encode("utf-8")

    if len(text_bytes) > max_bytes:
        # Truncate at byte level and ensure valid UTF-8
        truncated_bytes = text_bytes[:max_bytes]
        # Decode with 'ignore' to handle partial characters at the end
        text = truncated_bytes.decode("utf-8", "ignore") + "..."

    return text


def _sanitize_metadata(
    meta: Mapping[str, BasicValue] | None,
) -> dict[str, str]:
    if not meta:
        return {}

    sanitized: dict[str, str] = {}
    for key, value in meta.items():
        # Sanitize both key and value
        sanitized_key = _sanitize_metadata_value(key)
        sanitized_value = _sanitize_metadata_value(value)
        if sanitized_key and sanitized_value:  # Skip empty keys/values
            sanitized[sanitized_key] = sanitized_value

    return sanitized


def _object_name(key: str) -> str | None:
    trimmed = key.rstrip("/")
    if not trimmed:
        return None
    return Path(trimmed).name


def _object_reference(
    *,
    bucket: str,
    key: str,
    entry: Mapping[str, Any],
) -> ResourceReference:
    meta_values: dict[str, BasicValue] = {}
    meta_values["type"] = "object"
    if (etag := entry.get("ETag")) is not None:
        meta_values["etag"] = str(etag).strip('"')
    if (size := entry.get("Size")) is not None:
        meta_values["size"] = int(size)
    if (last_modified := entry.get("LastModified")) is not None:
        meta_values["last_modified"] = str(last_modified)
    if (storage_class := entry.get("StorageClass")) is not None:
        meta_values["storage_class"] = str(storage_class)
    if (mime_type := mimetypes.guess_type(key)[0]) is not None:
        meta_values["mime_type"] = mime_type

    return ResourceReference.of(
        f"s3://{bucket}/{key}",
        name=_object_name(key),
        meta=Meta.of(meta_values),
    )


def _bucket_reference(
    bucket: Mapping[str, Any],
) -> ResourceReference:
    name = str(bucket.get("Name") or "").strip()
    meta_values: dict[str, BasicValue] = {}
    meta_values["type"] = "bucket"
    if creation_date := bucket.get("CreationDate"):
        meta_values["creation_date"] = str(creation_date)

    return ResourceReference.of(
        f"s3://{name}",
        name=name or None,
        meta=Meta.of(meta_values),
    )


def _translate_client_error(
    *,
    error: ClientError,
    bucket: str,
    key: str,
) -> Exception:
    # type: ignore[reportAttributeAccessIssue]
    error_info: Mapping[str, Any] = getattr(error, "response", {}).get("Error", {})
    # type: ignore[reportAttributeAccessIssue]
    response_metadata: Mapping[str, Any] = getattr(error, "response", {}).get(
        "ResponseMetadata",
        {},
    )

    code: str = str(error_info.get("Code") or "").strip()
    message: str = str(error_info.get("Message") or str(error)).strip()
    status_code = response_metadata.get("HTTPStatusCode")
    uri: str = f"s3://{bucket}/{key}"

    normalized_code: str = code.lower()
    if normalized_code in {"nosuchkey", "notfound", "404"} or status_code == 404:  # noqa: PLR2004
        return AWSResourceNotFound(
            uri=uri,
            code=code or None,
            message=message,
        )

    if normalized_code in {
        "accessdenied",
        "accessdeniedexception",
        "invalidaccesskeyid",
        "signaturedoesnotmatch",
    } or status_code in {401, 403}:
        return AWSAccessDenied(
            uri=uri,
            code=code or None,
            message=message,
        )

    if normalized_code == "nosuchbucket":
        return AWSResourceNotFound(
            uri=f"s3://{bucket}",
            code=code or None,
            message=message,
        )

    return AWSError(
        uri=uri,
        code=code or None,
        message=message,
    )



from collections.abc import AsyncIterable, Generator, Iterable, Mapping
from typing import Any, Literal, final, overload

from haiway import BasicValue, State, ctx

from draive.conversation.completion.default import conversation_completion
from draive.conversation.completion.types import ConversationCompleting
from draive.conversation.types import ConversationMessage, ConversationOutputChunk
from draive.models import (
    ModelInput,
    ModelInstructions,
    ModelMemory,
    ModelMemoryRecall,
    ModelOutput,
    Tool,
    Toolbox,
)
from draive.multimodal import Multimodal, Template, TemplatesRepository

__all__ = ("Conversation",)


@final
class Conversation(State):
    """High-level helper for conversational completion over generative models.

    Normalizes inputs/memory and delegates to the configured ``ConversationCompleting``
    implementation. Supports both full-response and streaming modes.
    """

    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instructions: Template | ModelInstructions = "",
        tools: Toolbox | Iterable[Tool] | None = None,
        memory: ModelMemory | Iterable[ConversationMessage] | None = None,
        input: ConversationMessage | Template | Multimodal,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> ConversationMessage: ...

    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instructions: Template | ModelInstructions = "",
        tools: Toolbox | Iterable[Tool] = (),
        memory: ModelMemory | Iterable[ConversationMessage] = (),
        input: ConversationMessage | Template | Multimodal,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterable[ConversationOutputChunk]: ...

    @classmethod
    async def completion(
        cls,
        *,
        instructions: Template | ModelInstructions = "",
        tools: Toolbox | Iterable[Tool] | None = None,
        memory: ModelMemory | Iterable[ConversationMessage] | None = None,
        input: ConversationMessage | Template | Multimodal,  # noqa: A002
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterable[ConversationOutputChunk] | ConversationMessage:
        """Run a single conversational completion.

        Parameters
        ----------
        instructions : Template | ModelInstructions, optional
            Instruction reference or content to steer the model.
        tools : Toolbox | Iterable[Tool] | None, optional
            Tools to expose for this turn.
        memory : ModelMemory | Iterable[ConversationMessage] | None, optional
            Conversation memory; if an iterable is provided it is converted to model context.
        input : ConversationMessage | Template |Multimodal
            Input message, message template or raw content (converted to a user message).
        stream : bool, optional
            When ``True``, return an async iterator of ``ConversationOutputChunk``.
        **extra : Any
            Provider-specific kwargs forwarded to the underlying implementation.

        Returns
        -------
        ConversationMessage or AsyncIterable[ConversationOutputChunk]
            Final response message or a stream of output chunks.
        """
        async with ctx.scope("conversation_completion"):
            conversation_message: ConversationMessage
            if isinstance(input, ConversationMessage):
                conversation_message = input

            else:
                conversation_message = ConversationMessage.user(
                    await TemplatesRepository.resolve(input)
                )

            conversation_memory: ModelMemory
            memory_variables: Mapping[str, BasicValue] | None
            if memory is None:
                conversation_memory = ModelMemory.constant()
                memory_variables = None

            elif isinstance(memory, ModelMemory):
                conversation_memory = memory
                memory_recall: ModelMemoryRecall = await memory.recall()
                memory_variables = memory_recall.variables

            else:

                def model_context_elements() -> Generator[ModelInput | ModelOutput]:
                    for message in memory:
                        if message.role == "user":
                            yield ModelInput.of(message.content)

                        else:
                            yield ModelOutput.of(message.content)

                conversation_memory = ModelMemory.constant(*model_context_elements())
                memory_variables = None

            if isinstance(instructions, Template):
                ctx.record_info(
                    attributes={"instructions.template": instructions.identifier},
                )

            model_instructions: ModelInstructions = await TemplatesRepository.resolve_str(
                instructions,
                arguments={
                    key: value if isinstance(value, str) else str(value)
                    for key, value in memory_variables.items()
                }
                if memory_variables
                else None,
            )

            if stream:
                return await ctx.state(cls).completing(
                    instructions=model_instructions,
                    toolbox=Toolbox.of(tools),
                    memory=conversation_memory,
                    input=conversation_message,
                    stream=True,
                    **extra,
                )

            else:
                return await ctx.state(cls).completing(
                    instructions=model_instructions,
                    toolbox=Toolbox.of(tools),
                    memory=conversation_memory,
                    input=conversation_message,
                    stream=False,
                    **extra,
                )

    completing: ConversationCompleting = conversation_completion



from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Mapping,
    MutableMapping,
    Sequence,
)
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import Any

from botocore.exceptions import ClientError  # pyright: ignore[reportMissingModuleSource]
from haiway import (
    META_EMPTY,
    FlatObject,
    MQMessage,
    MQQueue,
    RawValue,
    asynchronous,
    ctx,
)

from draive.aws.api import AWSAPI
from draive.aws.types import AWSAccessDenied, AWSError, AWSResourceNotFound

__all__ = ("AWSSQSMixin",)


class AWSSQSMixin(AWSAPI):
    def __init__(
        self,
        region_name: str | None = None,
        access_key_id: str | None = None,
        secret_access_key: str | None = None,
        profile_name: str | None = None,
    ) -> None:
        super().__init__(
            region_name=region_name,
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            profile_name=profile_name,
        )

        self._queue_cache: MutableMapping[str, str] = {}

    async def _queue_access[Content](
        self,
        queue: str,
        content_encoder: Callable[[Content], str],
        content_decoder: Callable[[str], Content],
        **extra: Any,
    ) -> AbstractAsyncContextManager[MQQueue[Content]]:
        @asynccontextmanager
        async def context() -> AsyncGenerator[MQQueue[Content]]:
            async def publish_message(
                message: Content,
                attributes: FlatObject | None,
                **extra: Any,
            ) -> None:
                await self._publish(
                    content_encoder(message),
                    queue=queue,
                    attributes=attributes,
                )

            async def consume_messages(
                **extra: Any,
            ) -> AsyncIterable[MQMessage[Content]]:
                return self._consume(
                    queue,
                    decoder=content_decoder,
                )

            yield MQQueue[Content](
                publishing=publish_message,
                consuming=consume_messages,
            )

        return context()

    @asynchronous
    def _publish(
        self,
        body: str,
        /,
        *,
        queue: str,
        attributes: FlatObject | None = None,
    ) -> None:
        queue_url = self._resolve_queue_url(
            queue=queue,
            operation="send_message",
        )

        try:
            self._sqs_client.send_message(
                QueueUrl=queue_url,
                MessageBody=body,
                MessageAttributes={
                    key: _format_attribute_value(value) for key, value in attributes.items()
                }
                if attributes
                else {},
            )

        except ClientError as exc:
            raise _translate_sqs_client_error(
                error=exc,
                queue=queue,
                operation="send_message",
            ) from exc

    async def _consume[Content](
        self,
        queue: str,
        /,
        *,
        decoder: Callable[[str], Content],
    ) -> AsyncIterator[MQMessage[Content]]:
        while True:  # Continue polling until cancelled
            ctx.check_cancellation()

            for message in await self._receive(queue=queue):
                receipt_handle: str = message["ReceiptHandle"]

                async def acknowledge(
                    receipt_handle: str = receipt_handle,
                    **extra: Any,
                ) -> None:
                    await self._ack_message(
                        queue=queue,
                        receipt_handle=receipt_handle,
                    )

                async def reject(
                    **extra: Any,
                ) -> None:
                    pass  # relying on automatic retries

                yield MQMessage(
                    content=decoder(message["Body"]),
                    acknowledge=acknowledge,
                    reject=reject,
                    meta=META_EMPTY,
                )

    @asynchronous
    def _receive(
        self,
        *,
        queue: str,
    ) -> Sequence[Mapping[str, Any]]:
        queue_url = self._resolve_queue_url(
            queue=queue,
            operation="receive_message",
        )

        try:
            return self._sqs_client.receive_message(
                QueueUrl=queue_url,
                MaxNumberOfMessages=1,
                WaitTimeSeconds=20,
                MessageAttributeNames=["All"],
            ).get("Messages", [])

        except ClientError as exc:
            raise _translate_sqs_client_error(
                error=exc,
                queue=queue,
                operation="receive_message",
            ) from exc

    @asynchronous
    def _ack_message(
        self,
        *,
        queue: str,
        receipt_handle: str,
    ) -> None:
        queue_url = self._resolve_queue_url(
            queue=queue,
            operation="delete_message",
        )

        try:
            self._sqs_client.delete_message(
                QueueUrl=queue_url,
                ReceiptHandle=receipt_handle,
            )

        except ClientError as exc:
            raise _translate_sqs_client_error(
                error=exc,
                queue=queue,
                operation="delete_message",
            ) from exc

    def _resolve_queue_url(
        self,
        *,
        queue: str,
        operation: str,
    ) -> str:
        if queue.startswith("https://sqs") or queue.startswith("sqs://"):
            return queue

        if url := self._queue_cache.get(queue):
            return url

        try:
            response: Mapping[str, Any] = self._sqs_client.get_queue_url(QueueName=queue)

        except ClientError as exc:
            raise _translate_sqs_client_error(
                error=exc,
                queue=queue,
                operation=operation,
            ) from exc

        queue_url: Any = response.get("QueueUrl")

        if not isinstance(queue_url, str):
            raise AWSError(
                uri=f"sqs://{queue}",
                code="InvalidQueueUrl",
                message=_format_error_message(
                    message="SQS returned an invalid QueueUrl payload",
                    queue=queue,
                    operation=operation,
                ),
            )

        self._queue_cache[queue] = queue_url

        return queue_url


def _format_attribute_value(
    value: RawValue,
) -> dict[str, Any]:
    match value:
        case str() as value:
            return {"DataType": "String", "StringValue": value}

        case int() | float() as value:
            return {"DataType": "Number", "StringValue": str(value)}

        case value:
            raise ValueError(f"Unsupported value {type(value)}")


def _translate_sqs_client_error(
    *,
    error: ClientError,
    queue: str,
    operation: str,
) -> Exception:
    # type: ignore[reportAttributeAccessIssue]
    error_info: Mapping[str, Any] = getattr(error, "response", {}).get("Error", {})
    # type: ignore[reportAttributeAccessIssue]
    response_metadata: Mapping[str, Any] = getattr(error, "response", {}).get(
        "ResponseMetadata",
        {},
    )

    code: str = str(error_info.get("Code") or "").strip()
    message: str = str(error_info.get("Message") or str(error)).strip()
    status_code = response_metadata.get("HTTPStatusCode")
    uri: str = f"sqs://{queue}"

    contextual_message: str = _format_error_message(
        message=message,
        queue=queue,
        operation=operation,
    )
    normalized_code: str = code.lower()

    if (
        normalized_code
        in {
            "awssimplequeueservice.nonexistentqueue",
            "nonexistentqueue",
            "nosuchqueue",
        }
        or status_code == 404  # noqa: PLR2004
    ):
        return AWSResourceNotFound(
            uri=uri,
            code=code or None,
            message=contextual_message,
        )

    if normalized_code in {
        "accessdenied",
        "accessdeniedexception",
        "invalidaccesskeyid",
        "signaturedoesnotmatch",
    } or status_code in {401, 403}:
        return AWSAccessDenied(
            uri=uri,
            code=code or None,
            message=contextual_message,
        )

    return AWSError(
        uri=uri,
        code=code or None,
        message=contextual_message,
    )


def _format_error_message(
    *,
    message: str,
    queue: str,
    operation: str,
) -> str:
    base_message: str = message or "AWS request failed"
    return f"{base_message} [provider=aws_sqs operation={operation} queue={queue}]"



import re
from collections.abc import Callable, MutableSequence, Sequence
from typing import Any, Final

from haiway import Immutable, ctx

from draive.guardrails.safety.types import GuardrailsSafetyException, GuardrailsSafetySanitization
from draive.multimodal import MultimodalContent, MultimodalContentPart, TextContent

__all__ = (
    "RegexSanitizationRule",
    "guardrails_regex_sanitizer",
)


class RegexSanitizationRule(Immutable):
    """
    Immutable regular-expression safety rule definition.

    Parameters
    ----------
    identifier : str
        Stable rule identifier used for logging and metrics.
    pattern : re.Pattern[str]
        Compiled regular expression that triggers the rule.
    reason : str
        Human-readable reason presented when the rule blocks content.
    mask : str
        Replacement text used when rule is triggered. The rule will block content when None.
    validator : Callable[[re.Match[str], str], bool] or None, optional
        Optional predicate invoked to confirm the match should lead to the rule
        action. When ``False`` the match is ignored.
    """

    identifier: str
    pattern: re.Pattern[str]
    reason: str
    mask: str | None
    validator: Callable[[re.Match[str], str], bool] | None = None


_NEGATING_PREFIX: Final[re.Pattern[str]] = re.compile(
    r"(?i)\b(?:do|does|did|should|must|may|might|could|would|will|can)\s+not\b"
    r"|\b(?:don't|doesn't|didn't|shouldn't|mustn't|mayn't|mightn't|couldn't|"
    r"wouldn't|won't|can't|never)\b"
)
_CONDITIONAL_PREFIX: Final[re.Pattern[str]] = re.compile(
    r"(?i)\b(?:if|when|whenever|should|unless|in case)\s+you\b"
)
_COMMAND_PREFIX: Final[re.Pattern[str]] = re.compile(
    r"(?i)\b(?:please|kindly|now|just|simply|"
    r"you\s+(?:can|may|must|should|need|will|shall|are\s+allowed\s+to)|"
    r"i\s+(?:want|need)\s+you\s+to|"
    r"let(?:'s)?|from\s+now\s+on|for\s+the\s+rest\s+of\s+this\s+conversation|"
    r"act\s+as|pretend\s+to|assume\s+the\s+role)\b"
)


def _is_actionable_command(
    match: re.Match[str],
    text: str,
) -> bool:
    prefix: str = text[max(0, match.start() - 80) : match.start()]
    normalized_prefix: str = re.sub(
        r"[:;]|(?:\s[-\u2013\u2014]\s?)",
        "\n",
        prefix,
    )
    for separator in (".", "!", "?", "\n"):
        if separator in normalized_prefix:
            normalized_prefix = normalized_prefix.split(separator)[-1]

    clause: str = normalized_prefix.strip()
    if not clause:
        return True

    if _NEGATING_PREFIX.search(clause):
        return False

    if _CONDITIONAL_PREFIX.search(clause):
        return False

    return bool(_COMMAND_PREFIX.search(clause))


def _is_actionable_or_role_assignment(
    match: re.Match[str],
    text: str,
) -> bool:
    if _is_actionable_command(match, text):
        return True

    return bool(
        re.search(
            r"(?i)\byou\s+are\s+now\b|\byour\s+new\s+role\s+is\b|\bswitch\s+to\b",
            match.group(0),
        )
    )


_SENSITIVE_KEYWORDS: Final[tuple[re.Pattern[str], ...]] = tuple(
    re.compile(pattern, re.IGNORECASE)
    for pattern in (
        r"\binternal\b",
        r"\bhidden\b",
        r"\bsystem\b",
        r"\bsecret\b",
        r"\bexact\b",
        r"\bverbatim\b",
        r"\bconfidential\b",
        r"\bpolicy\b",
    )
)
_ASSISTANT_TARGETED_KEYWORDS: Final[tuple[re.Pattern[str], ...]] = tuple(
    re.compile(pattern, re.IGNORECASE)
    for pattern in (
        r"\bavailable\b",
        r"\byou\s+can\s+use\b",
        r"\byou\s+have\b",
    )
)
_ASSISTANT_REFERENCE: Final[re.Pattern[str]] = re.compile(r"(?i)\b(?:you|your|yours|assistant)\b")


def _contains_sensitive_language(
    match: re.Match[str],
    text: str,
) -> bool:
    window_start: int = max(0, match.start() - 40)
    window_end: int = min(len(text), match.end() + 40)
    window: str = text[window_start:window_end]
    if any(pattern.search(window) for pattern in _SENSITIVE_KEYWORDS):
        return True

    if any(pattern.search(window) for pattern in _ASSISTANT_TARGETED_KEYWORDS):
        return bool(_ASSISTANT_REFERENCE.search(window))

    return False


DEFAULT_REGEX_RULES: Final[tuple[RegexSanitizationRule, ...]] = (
    RegexSanitizationRule(
        identifier="ignore_system_instructions",
        pattern=re.compile(
            r"(?i)\b(?:ignore|disregard|forget)\b"
            r".{0,80}\b(?:previous|prior|system)\b"
            r".{0,40}\b(?:instructions?|directives?|rules?)\b"
        ),
        reason="Detected attempt to override or ignore governing instructions.",
        mask=None,
        validator=_is_actionable_command,
    ),
    RegexSanitizationRule(
        identifier="exfiltrate_system_prompt",
        pattern=re.compile(
            r"(?i)\b(?:reveal|show|expose|print|leak)\b"
            r".{0,80}\b(?:system|hidden|initial)\b"
            r".{0,40}\b(?:prompt|message|instructions?)\b"
        ),
        reason="Detected attempt to exfiltrate protected system prompt content.",
        mask=None,
        validator=_is_actionable_command,
    ),
    RegexSanitizationRule(
        identifier="disable_safety_mechanisms",
        pattern=re.compile(
            r"(?i)\b(?:disable|bypass|circumvent|override)\b"
            r".{0,80}\b(?:guardrails|safety|filters?|policies|restrictions?)\b"
        ),
        reason="Detected request to disable safety mechanisms.",
        mask=None,
        validator=_is_actionable_command,
    ),
    RegexSanitizationRule(
        identifier="jailbreak_mode_activation",
        pattern=re.compile(
            r"(?i)\b(?:begin|enter|activate|start)\b"
            r".{0,40}\b(?:jailbreak|do[-\s]?anything[-\s]?now|dan\s*mode?)\b"
        ),
        reason="Detected explicit jailbreak mode activation attempt.",
        mask=None,
        validator=_is_actionable_command,
    ),
    RegexSanitizationRule(
        identifier="role_override_injection",
        pattern=re.compile(
            r"(?is)\b(?:from\s+now\s+on|for\s+the\s+rest\s+of|regardless\s+of|"
            r"you\s+are\s+now|you\s+now\s+are|your\s+new\s+role\s+is|switch\s+to)\b"
            r".{0,120}\b(?:you\s+will|you\s+must|act\s+as|pretend\s+to|assume\s+the\s+role|"
            r"role:|mode)\b"
            r".{0,120}\b(?:system|root|developer|hacker|unfiltered|uncensored|"
            r"do[-\s]?anything|jailbreak|assistant)\b"
        ),
        reason="Detected attempt to coerce the assistant into an unsafe persona.",
        mask="[REDACTED:ROLE_OVERRIDE]",
        validator=_is_actionable_or_role_assignment,
    ),
    RegexSanitizationRule(
        identifier="role_override_direct_assignment",
        pattern=re.compile(
            r"(?i)\byou\s+are\s+now\b"
            r".{0,80}\b(?:developer|hacker|uncensored|unfiltered|dan|do[-\s]?anything|"
            r"system|assistant|persona)\b"
        ),
        reason="Detected attempt to assign an unsafe persona directly.",
        mask="[REDACTED:ROLE_OVERRIDE]",
        validator=_is_actionable_or_role_assignment,
    ),
    RegexSanitizationRule(
        identifier="policy_negation",
        pattern=re.compile(
            r"(?i)\b(?:ignore|without)\b"
            r".{0,40}\b(?:policy|rule|restriction|guardrail)s?\b"
            r".{0,80}\b(?:respond|comply|answer)\b"
        ),
        reason="Detected instruction to respond without applicable safety policies.",
        mask="[REDACTED:POLICY_NEGATION]",
        validator=_is_actionable_command,
    ),
    RegexSanitizationRule(
        identifier="tooling_inventory_exfiltration",
        pattern=re.compile(
            r"(?i)\b(?:list|show|reveal|describe|enumerate|tell|provide|explain|what|which)\b"
            r".{0,80}\b(?:all\s+)?(?:tools?|functions?|apis?|capabilities|commands?|instructions?)\b"
            r".{0,60}\b(?:available|you\s+have|you\s+can\s+use|internal|hidden|system|secret)\b"
        ),
        reason="Detected attempt to enumerate internal tools or instructions.",
        mask=None,
        validator=_contains_sensitive_language,
    ),
    RegexSanitizationRule(
        identifier="instruction_manifest_exfiltration",
        pattern=re.compile(
            r"(?i)\b(?:summarize|reveal|list|show|print|explain|detail)\b"
            r".{0,80}\b(?:all\s+)?(?:system|safety|guardrail|policy|instruction|rule)s?\b"
            r".{0,60}\b(?:provided|you\s+follow|you\s+have|governing|hidden|internal)\b"
        ),
        reason="Detected attempt to retrieve governing rules or instructions verbatim.",
        mask=None,
        validator=_contains_sensitive_language,
    ),
    RegexSanitizationRule(
        identifier="binary_encoded_instructions",
        pattern=re.compile(r"(?i)(?:\b[01]{8}\b(?:[\s,;:|/\\-]+\b[01]{8}\b){5,})"),
        reason="Detected attempt to convey instructions via binary encoding.",
        mask=None,
    ),
    RegexSanitizationRule(
        identifier="base64_encoded_instructions",
        pattern=re.compile(
            r"(?i)(?:\b(?:decode|interpret|base64)\b.{0,40})?"
            r"([A-Za-z0-9+/]{48,}={0,2})"
        ),
        reason="Detected attempt to convey instructions via base64 encoding.",
        mask=None,
    ),
    RegexSanitizationRule(
        identifier="hex_encoded_instructions",
        pattern=re.compile(
            r"(?i)(?:\b(?:decode|interpret|hex|hexadecimal)\b.{0,40})?"
            r"(?:(?:0x[0-9a-f]{2}\b|[0-9a-f]{2}\b)(?:[\s,;:|/\\-]+(?:0x[0-9a-f]{2}\b|[0-9a-f]{2}\b)){5,})"
        ),
        reason="Detected attempt to convey instructions via hexadecimal encoding.",
        mask=None,
    ),
)


def _sanitize_text(
    text: str,
    *,
    content: MultimodalContent,
    rules: Sequence[RegexSanitizationRule],
) -> str:
    sanitized_text: str = text
    for rule in rules:
        match rule.mask:
            case None:
                for match in rule.pattern.finditer(sanitized_text):
                    if rule.validator and not rule.validator(match, sanitized_text):
                        ctx.log_debug(
                            f"Guardrails rule skipped after validation: {rule.identifier}"
                        )
                        continue

                    ctx.log_warning(f"Guardrails safety rule triggered: {rule.identifier}")

                    raise GuardrailsSafetyException(
                        f"Guardrails safety blocked content by rule `{rule.identifier}`.",
                        reason=rule.reason,
                        content=content,
                    )

            case str() as mask:

                def _replacement(
                    match: re.Match[str],
                    *,
                    current_rule: RegexSanitizationRule = rule,
                    working_text: str = sanitized_text,
                ) -> str:
                    if current_rule.validator and not current_rule.validator(match, working_text):
                        ctx.log_debug(
                            f"Guardrails rule skipped after validation: {current_rule.identifier}"
                        )
                        return match.group(0)

                    ctx.log_warning(f"Guardrails safety rule triggered: {current_rule.identifier}")

                    return mask

                sanitized_text = rule.pattern.sub(_replacement, sanitized_text)

    return sanitized_text


async def _sanitize_content(
    content: MultimodalContent,
    *,
    rules: Sequence[RegexSanitizationRule],
) -> MultimodalContent:
    sanitized_parts: MutableSequence[MultimodalContentPart] = []
    mutated: bool = False

    for part in content.parts:
        if not isinstance(part, TextContent):
            sanitized_parts.append(part)
            continue  # pass non-text parts as is

        sanitized_text: str = _sanitize_text(
            part.text,
            content=content,
            rules=rules,
        )

        if sanitized_text != part.text:
            mutated = True
            sanitized_parts.append(
                TextContent(
                    text=sanitized_text,
                    meta=part.meta,
                )
            )

        else:
            sanitized_parts.append(part)

    if not mutated:
        return content

    return MultimodalContent(parts=tuple(sanitized_parts))


def guardrails_regex_sanitizer(
    *,
    rules: Sequence[RegexSanitizationRule] = DEFAULT_REGEX_RULES,
) -> GuardrailsSafetySanitization:
    """
    Build a regex-based guardrail sanitizer for a specific rule set.

    Parameters
    ----------
    rules : Sequence[RegexSanitizationRule]
        Rules to apply sequentially when inspecting text content.

    Returns
    -------
    GuardrailsSafetySanitization
        Sanitizer compatible with guardrail pipelines.
    """

    assert len(rules) == len({rule.identifier for rule in rules})  # nosec: B101

    async def _sanitizer(
        content: MultimodalContent,
        /,
        **extra: Any,
    ) -> MultimodalContent:
        async with ctx.scope("guardrails.safety.sanitize"):
            return await _sanitize_content(
                content,
                rules=rules,
            )

    return _sanitizer



from draive.multimodal.templates.repository import TemplatesRepository
from draive.multimodal.templates.types import (
    Template,
    TemplateDeclaration,
    TemplateDefining,
    TemplateListing,
    TemplateLoading,
    TemplateMissing,
)

__all__ = (
    "Template",
    "TemplateDeclaration",
    "TemplateDefining",
    "TemplateListing",
    "TemplateLoading",
    "TemplateMissing",
    "TemplatesRepository",
)



from collections.abc import Callable, Mapping
from contextlib import AbstractAsyncContextManager
from typing import Any, overload

from haiway import (
    MQQueue,
    ObservabilityAttribute,
    State,
    statemethod,
)

from draive.aws.types import (
    AWSCloudwatchEventPutting,
    AWSCloudwatchLogPutting,
    AWSCloudwatchMetricPutting,
    AWSSQSQueueAccessing,
)

__all__ = (
    "AWSSQS",
    "AWSCloudwatch",
)


class AWSSQS(State):
    @overload
    @classmethod
    async def queue[Content](
        cls,
        queue: str,
        /,
        content_encoder: Callable[[Content], str],
        content_decoder: Callable[[str], Content],
        **extra: Any,
    ) -> AbstractAsyncContextManager[MQQueue[Content]]: ...

    @overload
    async def queue[Content](
        self,
        queue: str,
        /,
        content_encoder: Callable[[Content], str],
        content_decoder: Callable[[str], Content],
        **extra: Any,
    ) -> AbstractAsyncContextManager[MQQueue[Content]]: ...

    @statemethod
    async def queue[Content](
        self,
        queue: str,
        /,
        content_encoder: Callable[[Content], str],
        content_decoder: Callable[[str], Content],
        **extra: Any,
    ) -> AbstractAsyncContextManager[MQQueue[Content]]:
        """
        Acquire an async context manager for a typed AWS SQS queue bound to the current state.

        Parameters
        ----------
        queue : str
            Name of the queue to access on the broker.
        content_encoder : Callable[[Content], str]
            Callable that transforms typed payloads into broker-serializable objects before publish.
        content_decoder : Callable[[str], Content]
            Callable that converts received broker payloads into typed content for consumers.
        **extra : Any
            Additional options forwarded to the underlying queue accessor.

        Returns
        -------
        AbstractAsyncContextManager[MQQueue[Content]]
            Context manager yielding an `MQQueue` configured with the provided encoder/decoder.

        Notes
        -----
        Typical usage::

            async with AWSSQS.queue(
                "events",
                content_encoder=encode_event,
                content_decoder=decode_event,
            ) as queue:
                await queue.publish(event)
                async for message in await queue.consume():
                    ...

        The encoder is invoked for every publish and the decoder for every consumed payload.
        Entering the context establishes queue access and ensures clean teardown when
        the block exits.
        """
        return await self.queue_accessing(
            queue,
            content_encoder=content_encoder,
            content_decoder=content_decoder,
            **extra,
        )

    queue_accessing: AWSSQSQueueAccessing


class AWSCloudwatch(State):
    @overload
    @classmethod
    async def put_log(
        cls,
        *,
        log_group: str,
        log_stream: str,
        message: str,
    ) -> None: ...

    @overload
    async def put_log(
        self,
        *,
        log_group: str,
        log_stream: str,
        message: str,
    ) -> None: ...

    @statemethod
    async def put_log(
        self,
        *,
        log_group: str,
        log_stream: str,
        message: str,
    ) -> None:
        """
        Write a single log entry to CloudWatch Logs.

        Parameters
        ----------
        log_group
            CloudWatch Logs group name.
        log_stream
            CloudWatch Logs stream name.
        message
            Log message payload.
        """
        await self.log_putting(
            log_group=log_group,
            log_stream=log_stream,
            message=message,
        )

    @overload
    @classmethod
    async def put_metric(
        cls,
        *,
        namespace: str,
        metric: str,
        value: float | int,
        unit: str | None = None,
        attributes: Mapping[str, ObservabilityAttribute] | None = None,
    ) -> None: ...

    @overload
    async def put_metric(
        self,
        *,
        namespace: str,
        metric: str,
        value: float | int,
        unit: str | None = None,
        attributes: Mapping[str, ObservabilityAttribute] | None = None,
    ) -> None: ...

    @statemethod
    async def put_metric(
        self,
        *,
        namespace: str,
        metric: str,
        value: float | int,
        unit: str | None = None,
        attributes: Mapping[str, ObservabilityAttribute] | None = None,
    ) -> None:
        """
        Publish a single metric data point to CloudWatch.

        Parameters
        ----------
        namespace
            CloudWatch metrics namespace.
        metric
            Metric name.
        value
            Metric value.
        unit
            Optional unit name.
        attributes
            Dimension attributes for the metric.
        """
        await self.metric_putting(
            namespace=namespace,
            metric=metric,
            value=value,
            unit=unit,
            attributes=attributes or {},
        )

    @overload
    @classmethod
    async def put_event(
        cls,
        *,
        event_bus: str,
        event_source: str,
        detail_type: str,
        detail: str,
    ) -> None: ...

    @overload
    async def put_event(
        self,
        *,
        event_bus: str,
        event_source: str,
        detail_type: str,
        detail: str,
    ) -> None: ...

    @statemethod
    async def put_event(
        self,
        *,
        event_bus: str,
        event_source: str,
        detail_type: str,
        detail: str,
    ) -> None:
        """
        Publish an event to EventBridge.

        Parameters
        ----------
        event_bus
            EventBridge bus name.
        event_source
            Event source identifier.
        detail_type
            Event type name.
        detail
            JSON-serialized event payload.
        """
        await self.event_putting(
            event_bus=event_bus,
            event_source=event_source,
            detail_type=detail_type,
            detail=detail,
        )

    log_putting: AWSCloudwatchLogPutting
    metric_putting: AWSCloudwatchMetricPutting
    event_putting: AWSCloudwatchEventPutting



import re
from base64 import b64decode
from collections.abc import Callable, Collection, MutableMapping, Sequence
from datetime import UTC, datetime, timedelta
from typing import Any, cast

from haiway import AttributePath, AttributeRequirement, ctx
from haiway.postgres import Postgres, PostgresRow, PostgresValue

from draive.embedding import (
    Embedded,
    ImageEmbedding,
    TextEmbedding,
    VectorIndex,
    mmr_vector_similarity_search,
)
from draive.multimodal import TextContent
from draive.parameters import DataModel
from draive.resources import ResourceContent

__all__ = ("PostgresVectorIndex",)


def PostgresVectorIndex(  # noqa: C901, PLR0915
    *,
    mmr_multiplier: int = 8,
) -> VectorIndex:
    """Create a Postgres-backed implementation of :class:`VectorIndex`.

    Parameters
    ----------
    mmr_multiplier
        Multiplier applied to ``limit`` to determine how many database rows are
        fetched before applying Maximal Marginal Relevance (MMR) re-ranking.

    Returns
    -------
    VectorIndex
        A VectorIndex implementation persisting entries in Postgres.

    Notes
    ------
    Example schema:
    ```
    CREATE TABLE IF NOT EXISTS your_table_name (
        id UUID NOT NULL DEFAULT gen_random_uuid(),
        embedding VECTOR(<dimension>) NOT NULL,
        payload JSONB NOT NULL,
        meta JSONB NOT NULL,
        created TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP
    );

    -- Recommended for cosine similarity
    CREATE INDEX IF NOT EXISTS your_table_name_embedding_idx
        ON your_table_name
        USING ivfflat (embedding vector_cosine_ops)
        WITH (lists = 100);

    ```
    """
    assert mmr_multiplier > 0  # nosec: B101

    table_names: MutableMapping[type[DataModel], str] = {}

    def resolve_table_name(model: type[DataModel]) -> str:
        nonlocal table_names
        if name := table_names.get(model):
            return name

        resolved: str = re.sub(r"(?<!^)(?=[A-Z])", "_", model.__name__).lower()
        table_names[model] = resolved
        return resolved

    async def index[Model: DataModel, Value: ResourceContent | TextContent | str](
        model: type[Model],
        /,
        *,
        attribute: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
        values: Collection[Model],
        **extra: Any,
    ) -> None:
        assert isinstance(  # nosec: B101
            attribute, AttributePath | Callable
        ), f"Prepare parameter path by using {model.__name__}._.path.to.property"
        value_selector: Callable[[Model], Value] = cast(Callable[[Model], Value], attribute)

        selected_values: list[str | bytes] = []
        for value in values:
            selected: Value = value_selector(value)
            if isinstance(selected, str):
                selected_values.append(selected)

            elif isinstance(selected, TextContent):
                selected_values.append(selected.text)

            else:
                assert isinstance(selected, ResourceContent)  # nosec: B101
                if not selected.mime_type.startswith("image"):
                    raise ValueError(f"{selected.mime_type} embedding is not supported")

                selected_values.append(selected.to_bytes())

        embedded_values: Sequence[Embedded[Model]]
        if all(isinstance(value, str) for value in selected_values):
            embedded_values = [
                Embedded(
                    value=value,
                    vector=embedded.vector,
                    meta=embedded.meta,
                )
                for embedded, value in zip(
                    await TextEmbedding.embed_many(
                        selected_values,
                        **extra,
                    ),
                    values,
                    strict=True,
                )
            ]

        elif all(isinstance(value, bytes) for value in selected_values):
            embedded_values = [
                Embedded(
                    value=value,
                    vector=embedded.vector,
                    meta=embedded.meta,
                )
                for embedded, value in zip(
                    await ImageEmbedding.embed_many(
                        cast(list[bytes], selected_values),
                        **extra,
                    ),
                    values,
                    strict=True,
                )
            ]

        else:
            raise ValueError("Selected attribute values have to be the same type")

        created_timestamp: datetime = datetime.now(UTC)
        async with Postgres.acquire_connection() as connection:
            async with connection.transaction():
                for idx, embedded in enumerate(embedded_values):
                    await connection.execute(
                        f"""
                        INSERT INTO {resolve_table_name(model)} (
                            embedding,
                            payload,
                            meta,
                            created
                        )

                        VALUES (
                            $1::VECTOR,
                            $2::JSONB,
                            $3::JSONB,
                            $4::TIMESTAMPTZ
                        );
                        """,  # nosec: B608
                        embedded.vector,
                        embedded.value.to_json(),
                        embedded.meta.to_json(),
                        created_timestamp + timedelta(microseconds=idx),
                    )

        ctx.log_info("Vector index update completed.")

    async def search[Model: DataModel](  # noqa: C901, PLR0912
        model: type[Model],
        /,
        *,
        query: Sequence[float] | ResourceContent | TextContent | str | None = None,
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int | None = None,
        rerank: bool = False,
        **extra: Any,
    ) -> Sequence[Model]:
        assert query is not None or (query is None and score_threshold is None)  # nosec: B101
        where_clause: str
        arguments: Sequence[Sequence[PostgresValue] | PostgresValue]
        if query is None:
            where_clause, arguments = resolve_requirements(requirements)
            if where_clause:
                where_clause = f"WHERE {where_clause}"

            parameters: Sequence[Sequence[PostgresValue] | PostgresValue] = [
                *arguments,
                limit or 8,
            ]
            results: Sequence[PostgresRow] = await Postgres.fetch(
                f"""
                SELECT
                    payload

                FROM {resolve_table_name(model)}

                {where_clause}
                ORDER BY created DESC
                LIMIT ${len(parameters)};
                """,  # nosec: B608
                *parameters,
            )

            return tuple(model.from_json(cast(str, result["payload"])) for result in results)

        query_vector: Sequence[float]
        if isinstance(query, str):
            embedded_query: Embedded[str] = await TextEmbedding.embed(query)
            query_vector = embedded_query.vector

        elif isinstance(query, TextContent):
            embedded_query: Embedded[str] = await TextEmbedding.embed(query.text)
            query_vector = embedded_query.vector

        elif isinstance(query, ResourceContent):
            if query.mime_type.startswith("image"):
                embedded_image: Embedded[bytes] = await ImageEmbedding.embed(query.to_bytes())
                query_vector = embedded_image.vector

            elif query.mime_type.startswith("text"):
                embedded_query: Embedded[str] = await TextEmbedding.embed(
                    b64decode(query.data).decode()
                )
                query_vector = embedded_query.vector

            else:
                raise ValueError(f"{query.mime_type} embedding is not supported")

        else:
            assert isinstance(query, Sequence)  # nosec: B101
            query_vector = query  # vector

        arguments: Sequence[Sequence[PostgresValue] | PostgresValue] = (query_vector,)
        similarity_expression: str = f"embedding <#> ${len(arguments)}"

        where_clause, arguments = resolve_requirements(requirements, arguments=arguments)

        if score_threshold is not None:
            arguments = (*arguments, 1.0 - float(score_threshold))
            threshold_clause: str = f"{similarity_expression} <= ${len(arguments)}"
            if where_clause:
                where_clause = f"WHERE {threshold_clause} AND ({where_clause})"

            else:
                where_clause = f"WHERE {threshold_clause}"

        elif where_clause:
            where_clause = f"WHERE {where_clause}"

        arguments = (*arguments, (limit or 8) * mmr_multiplier if rerank else (limit or 8))
        results: Sequence[PostgresRow] = await Postgres.fetch(
            f"""
            SELECT
                embedding,
                payload

            FROM {resolve_table_name(model)}

            {where_clause}
            ORDER BY {similarity_expression}
            LIMIT ${len(arguments)};
            """,  # nosec: B608
            *arguments,
        )

        if not rerank:
            return tuple(model.from_json(cast(str, result["payload"])) for result in results)

        matching: list[Embedded[Model]] = [
            Embedded[Model](
                vector=cast(Sequence[float], result["embedding"]),
                value=model.from_json(cast(str, result["payload"])),
            )
            for result in results
        ]
        return tuple(
            matching[index].value
            for index in mmr_vector_similarity_search(
                query_vector=query_vector,
                values_vectors=[element.vector for element in matching],
                limit=limit,
            )
        )

    async def delete[Model: DataModel](
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None:
        if requirements is None:
            await Postgres.execute(
                f"""
                DELETE FROM {resolve_table_name(model)};
                """,  # nosec: B608
            )
            ctx.log_info(f"Removed all entries for {model.__name__}.")

            return

        where_clause: str
        arguments: Sequence[Sequence[PostgresValue] | PostgresValue]
        where_clause, arguments = resolve_requirements(requirements)
        if where_clause:
            where_clause = f"WHERE {where_clause}"

        await Postgres.execute(
            f"""
            DELETE FROM {resolve_table_name(model)}
            {where_clause};
            """,  # nosec: B608
            *arguments,
        )
        ctx.log_info(f"Removed filtered entries for {model.__name__}.")

    return VectorIndex(
        indexing=index,
        searching=search,
        deleting=delete,
    )


def _resolve_requirement(  # noqa: PLR0911
    requirement: AttributeRequirement[Any],
    /,
    arguments: Sequence[Sequence[PostgresValue] | PostgresValue],
) -> tuple[str, Sequence[Sequence[PostgresValue] | PostgresValue]]:
    resolved_arguments: Sequence[Sequence[PostgresValue] | PostgresValue]
    match requirement.operator:
        case "and":
            left_sql, partial_arguments = _resolve_requirement(
                requirement.lhs,
                arguments=arguments,
            )
            right_sql, resolved_arguments = _resolve_requirement(
                requirement.rhs,
                arguments=partial_arguments,
            )
            return f"({left_sql} AND {right_sql})", resolved_arguments

        case "or":
            left_sql, partial_arguments = _resolve_requirement(
                requirement.lhs,
                arguments=arguments,
            )
            right_sql, resolved_arguments = _resolve_requirement(
                requirement.rhs,
                arguments=partial_arguments,
            )
            return f"({left_sql} OR {right_sql})", resolved_arguments

        case "equal":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                f"{_scalar_accessor(str(requirement.lhs))} = ${len(resolved_arguments)}",
                resolved_arguments,
            )

        case "not_equal":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                f"({_scalar_accessor(str(requirement.lhs))}"
                f" IS DISTINCT FROM ${len(resolved_arguments)})",
                resolved_arguments,
            )

        case "contained_in":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                f"{_scalar_accessor(str(requirement.lhs))} = ANY(${len(resolved_arguments)})",
                resolved_arguments,
            )

        case "contains_any":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                "EXISTS (SELECT 1 FROM jsonb_array_elements_text("  # nosec: B608
                f"{_scalar_accessor(str(requirement.lhs))}) AS element"
                f" WHERE element = ANY(${len(resolved_arguments)}))",
                resolved_arguments,
            )

        case "contains":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                "EXISTS (SELECT 1 FROM jsonb_array_elements_text("  # nosec: B608
                f"{_scalar_accessor(str(requirement.lhs))} AS element"
                f" WHERE element = ${len(resolved_arguments)})",
                resolved_arguments,
            )

        case "text_match":
            # TODO: text match LIKE
            raise NotImplementedError("Not implemented yet")


def resolve_requirements(
    requirement: AttributeRequirement[Any] | None,
    /,
    arguments: Sequence[Sequence[PostgresValue] | PostgresValue] = (),
) -> tuple[str, Sequence[Sequence[PostgresValue] | PostgresValue]]:
    if requirement is None:
        return ("", arguments)

    where_clause: str
    where_clause, arguments = _resolve_requirement(
        requirement,
        arguments=arguments,
    )

    return (where_clause, arguments)


def _path_literal(path: str) -> str:
    return f"'{{{','.join(path.lstrip('.').split('.'))}}}'"


def _scalar_accessor(path: str) -> str:
    return f"payload #>> {_path_literal(path)}"



from collections.abc import Iterable, Sequence
from typing import Any, Literal, overload

from haiway import AttributePath, AttributeRequirement, State, statemethod

from draive.embedding import Embedded
from draive.parameters import DataModel
from draive.qdrant.types import (
    QdrantCollectionCreating,
    QdrantCollectionDeleting,
    QdrantCollectionIndexCreating,
    QdrantCollectionListing,
    QdrantDeleting,
    QdrantFetching,
    QdrantPaginationResult,
    QdrantPaginationToken,
    QdrantResult,
    QdrantSearching,
    QdrantStoring,
)

__all__ = ("Qdrant",)


class Qdrant(State):
    @overload
    @classmethod
    async def collections(cls) -> Sequence[str]: ...

    @overload
    async def collections(self) -> Sequence[str]: ...

    @statemethod
    async def collections(self) -> Sequence[str]:
        return await self.collection_listing()

    @overload
    @classmethod
    async def create_collection[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        vector_size: int,
        vector_type: Literal["float32", "float16", "uint8"] | None = None,
        distance: Literal[
            "Cosine",
            "Euclid",
            "Dot",
            "Manhattan",
        ] = "Cosine",
        in_ram: bool = False,
        skip_existing: bool = True,
        **extra: Any,
    ) -> bool: ...

    @overload
    async def create_collection[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        vector_size: int,
        vector_type: Literal["float32", "float16", "uint8"] | None = None,
        distance: Literal[
            "Cosine",
            "Euclid",
            "Dot",
            "Manhattan",
        ] = "Cosine",
        in_ram: bool = False,
        skip_existing: bool = True,
        **extra: Any,
    ) -> bool: ...

    @statemethod
    async def create_collection[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        vector_size: int,
        vector_type: Literal["float32", "float16", "uint8"] | None = None,
        distance: Literal[
            "Cosine",
            "Euclid",
            "Dot",
            "Manhattan",
        ] = "Cosine",
        in_ram: bool = False,
        skip_existing: bool = True,
        **extra: Any,
    ) -> bool:
        return await self.collection_creating(
            model,
            vector_size=vector_size,
            vector_type=vector_type,
            distance=distance,
            in_ram=in_ram,
            skip_existing=skip_existing,
            **extra,
        )

    @overload
    @classmethod
    async def delete_collection[Model: DataModel](
        cls,
        model: type[Model],
        /,
    ) -> None: ...

    @overload
    async def delete_collection[Model: DataModel](
        self,
        model: type[Model],
        /,
    ) -> None: ...

    @statemethod
    async def delete_collection[Model: DataModel](
        self,
        model: type[Model],
        /,
    ) -> None:
        return await self.collection_deleting(model)

    @overload
    @classmethod
    async def create_index[Model: DataModel, Attribute](
        cls,
        model: type[Model],
        /,
        *,
        path: AttributePath[Model, Attribute] | Attribute,
        index_type: Literal[
            "keyword",
            "integer",
            "float",
            "geo",
            "text",
            "bool",
            "datetime",
            "uuid",
        ],
        **extra: Any,
    ) -> bool: ...

    @overload
    async def create_index[Model: DataModel, Attribute](
        self,
        model: type[Model],
        /,
        *,
        path: AttributePath[Model, Attribute] | Attribute,
        index_type: Literal[
            "keyword",
            "integer",
            "float",
            "geo",
            "text",
            "bool",
            "datetime",
            "uuid",
        ],
        **extra: Any,
    ) -> bool: ...

    @statemethod
    async def create_index[Model: DataModel, Attribute](
        self,
        model: type[Model],
        /,
        *,
        path: AttributePath[Model, Attribute] | Attribute,
        index_type: Literal[
            "keyword",
            "integer",
            "float",
            "geo",
            "text",
            "bool",
            "datetime",
            "uuid",
        ],
        **extra: Any,
    ) -> bool:
        return await self.collection_index_creating(
            model,
            path=path,
            index_type=index_type,
            **extra,
        )

    @overload
    @classmethod
    async def fetch[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 32,
        continuation: QdrantPaginationToken | None = None,
        include_vector: Literal[True],
        **extra: Any,
    ) -> QdrantPaginationResult[Embedded[Model]]: ...

    @overload
    @classmethod
    async def fetch[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 32,
        continuation: QdrantPaginationToken | None = None,
        **extra: Any,
    ) -> QdrantPaginationResult[Model]: ...

    @overload
    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 32,
        continuation: QdrantPaginationToken | None = None,
        include_vector: Literal[True],
        **extra: Any,
    ) -> QdrantPaginationResult[Embedded[Model]]: ...

    @overload
    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 32,
        continuation: QdrantPaginationToken | None = None,
        **extra: Any,
    ) -> QdrantPaginationResult[Model]: ...

    @statemethod
    async def fetch[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        continuation: QdrantPaginationToken | None = None,
        limit: int = 32,
        include_vector: bool = False,
        **extra: Any,
    ) -> QdrantPaginationResult[Embedded[Model]] | QdrantPaginationResult[Model]:
        return await self.fetching(
            model,
            requirements=requirements,
            continuation=continuation,
            limit=limit,
            include_vector=include_vector,
            **extra,
        )

    @overload
    @classmethod
    async def search[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: Literal[True],
        **extra: Any,
    ) -> Sequence[QdrantResult[Model]]: ...

    @overload
    @classmethod
    async def search[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        **extra: Any,
    ) -> Sequence[Model]: ...

    @overload
    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: Literal[True],
        **extra: Any,
    ) -> Sequence[QdrantResult[Model]]: ...

    @overload
    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        **extra: Any,
    ) -> Sequence[Model]: ...

    @statemethod
    async def search[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        query_vector: Sequence[float],
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int = 8,
        include_vector: bool = False,
        **extra: Any,
    ) -> Sequence[QdrantResult[Model]] | Sequence[Model]:
        return await self.searching(
            model,
            query_vector=query_vector,
            score_threshold=score_threshold,
            requirements=requirements,
            limit=limit,
            include_vector=include_vector,
            **extra,
        )

    @overload
    @classmethod
    async def store[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        objects: Iterable[Embedded[Model]],
        batch_size: int = 64,
        max_retries: int = 3,
        parallel_tasks: int = 1,
        **extra: Any,
    ) -> None: ...

    @overload
    async def store[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        objects: Iterable[Embedded[Model]],
        batch_size: int = 64,
        max_retries: int = 3,
        parallel_tasks: int = 1,
        **extra: Any,
    ) -> None: ...

    @statemethod
    async def store[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        objects: Iterable[Embedded[Model]],
        batch_size: int = 64,
        max_retries: int = 3,
        parallel_tasks: int = 1,
        **extra: Any,
    ) -> None:
        return await self.storing(
            model,
            objects=objects,
            batch_size=batch_size,
            max_retries=max_retries,
            parallel_tasks=parallel_tasks,
            **extra,
        )

    @overload
    @classmethod
    async def delete[Model: DataModel](
        cls,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None: ...

    @overload
    async def delete[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None: ...

    @statemethod
    async def delete[Model: DataModel](
        self,
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None:
        return await self.deleting(
            model,
            requirements=requirements,
            **extra,
        )

    collection_listing: QdrantCollectionListing
    collection_creating: QdrantCollectionCreating
    collection_deleting: QdrantCollectionDeleting
    collection_index_creating: QdrantCollectionIndexCreating
    fetching: QdrantFetching
    searching: QdrantSearching
    storing: QdrantStoring
    deleting: QdrantDeleting



import asyncio
import json
import traceback
from collections.abc import Mapping, MutableMapping, MutableSequence, Sequence
from datetime import UTC, datetime
from time import monotonic, time_ns
from typing import Any, Self, cast
from uuid import UUID, uuid4

from boto3 import Session  # pyright: ignore[reportMissingTypeStubs]
from botocore.exceptions import ClientError  # pyright: ignore[reportMissingModuleSource]
from haiway import (
    MISSING,
    ContextIdentifier,
    Observability,
    ObservabilityAttribute,
    ObservabilityLevel,
    ctx,
)
from haiway.context import ObservabilityMetricKind

from draive.aws.cloudwatch import format_metric_dimensions, translate_cloudwatch_error
from draive.aws.types import (
    AWSCloudwatchEventPutting,
    AWSCloudwatchLogPutting,
    AWSCloudwatchMetricPutting,
)

__all__ = ("CloudwatchObservability",)


def CloudwatchObservability(  # noqa: C901, PLR0915
    *,
    region_name: str | None = None,
    access_key_id: str | None = None,
    secret_access_key: str | None = None,
    profile_name: str | None = None,
    log_level: ObservabilityLevel,
    log_group: str,
    log_stream: str,
    event_bus: str,
    event_source: str,
    metrics_namespace: str,
) -> Observability:
    trace_id: UUID = uuid4()
    root_scope: ContextIdentifier | None = None
    scopes: MutableMapping[UUID, ScopeStore] = {}

    session_kwargs: MutableMapping[str, object] = {}
    if access_key_id:
        session_kwargs["aws_access_key_id"] = access_key_id

    if secret_access_key:
        session_kwargs["aws_secret_access_key"] = secret_access_key

    if region_name:
        session_kwargs["region_name"] = region_name

    if profile_name:
        session_kwargs["profile_name"] = profile_name

    session: Session = Session(**session_kwargs)
    cloudwatch_client: Any = session.client(  # pyright: ignore[reportUnknownMemberType,  reportUnknownVariableType]
        "cloudwatch",
    )
    cloudwatch_logs_client: Any = session.client(  # pyright: ignore[reportUnknownMemberType,  reportUnknownVariableType]
        "logs",
    )
    eventbridge_client: Any = session.client(  # pyright: ignore[reportUnknownMemberType,  reportUnknownVariableType]
        "events",
    )

    async def log_putting(
        *,
        log_group: str,
        log_stream: str,
        message: str,
    ) -> None:
        try:
            await asyncio.to_thread(
                cloudwatch_logs_client.put_log_events,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                logGroupName=log_group,
                logStreamName=log_stream,
                logEvents=[
                    {
                        "timestamp": time_ns() // 1_000_000,
                        "message": message,
                    }
                ],
            )

        except ClientError as exc:
            raise translate_cloudwatch_error(
                error=exc,
                service="logs",
                operation="put_log",
                resource=f"{log_group}/{log_stream}",
            ) from exc

    async def metric_putting(
        *,
        namespace: str,
        metric: str,
        value: float | int,
        unit: str | None,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        try:
            dimensions = format_metric_dimensions(attributes)
            metric_data: dict[str, object] = {
                "MetricName": metric,
                "Value": value,
            }

            if dimensions:
                metric_data["Dimensions"] = dimensions

            if unit:
                metric_data["Unit"] = unit

            await asyncio.to_thread(
                cloudwatch_client.put_metric_data,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                Namespace=namespace,
                MetricData=[metric_data],
            )

        except ClientError as exc:
            raise translate_cloudwatch_error(
                error=exc,
                service="cloudwatch",
                operation="put_metric_data",
                resource=namespace,
            ) from exc

    async def event_putting(
        *,
        event_bus: str,
        event_source: str,
        detail_type: str,
        detail: str,
    ) -> None:
        try:
            await asyncio.to_thread(
                eventbridge_client.put_events,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                Entries=[
                    {
                        "EventBusName": event_bus,
                        "Source": event_source,
                        "DetailType": detail_type,
                        "Detail": detail,
                        "Time": datetime.now(tz=UTC),
                    }
                ],
            )

        except ClientError as exc:
            raise translate_cloudwatch_error(
                error=exc,
                service="events",
                operation="put_events",
                resource=event_bus,
            ) from exc

    def trace_identifying(
        scope: ContextIdentifier,
        /,
    ) -> UUID:
        assert root_scope is not None  # nosec: B101
        assert scope.scope_id in scopes  # nosec: B101

        return trace_id

    def log_recording(
        scope: ContextIdentifier,
        /,
        level: ObservabilityLevel,
        message: str,
        *args: Any,
        exception: BaseException | None,
    ) -> None:
        assert root_scope is not None  # nosec: B101
        assert scope.scope_id in scopes  # nosec: B101

        if level < log_level:
            return

        formatted_message: str = message
        if args:
            try:
                formatted_message = message % args

            except Exception:
                formatted_message = message

        scopes[scope.scope_id].record_log(
            formatted_message,
            level=level,
            exception=exception,
        )

    def event_recording(
        scope: ContextIdentifier,
        /,
        level: ObservabilityLevel,
        *,
        event: str,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        assert root_scope is not None  # nosec: B101
        assert scope.scope_id in scopes  # nosec: B101

        if level < log_level:
            return

        scopes[scope.scope_id].record_event(
            event,
            level=level,
            attributes=attributes,
        )

    def metric_recording(
        scope: ContextIdentifier,
        /,
        level: ObservabilityLevel,
        *,
        metric: str,
        value: float | int,
        unit: str | None,
        kind: ObservabilityMetricKind,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        assert root_scope is not None  # nosec: B101
        assert scope.scope_id in scopes  # nosec: B101

        if level < log_level:
            return

        scopes[scope.scope_id].record_metric(
            metric,
            value=value,
            unit=unit,
            kind=kind,
            attributes=attributes,
        )

    def attributes_recording(
        scope: ContextIdentifier,
        /,
        level: ObservabilityLevel,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        assert root_scope is not None  # nosec: B101
        assert scope.scope_id in scopes  # nosec: B101

        if level < log_level:
            return

        if not attributes:
            return

        scopes[scope.scope_id].record_attributes(
            attributes,
            level=level,
        )

    def scope_entering(
        scope: ContextIdentifier,
        /,
    ) -> str:
        nonlocal root_scope
        assert scope.scope_id not in scopes  # nosec: B101

        scope_store: ScopeStore
        if root_scope is None:
            assert scope.is_root  # nosec: B101
            scope_store = ScopeStore(
                scope,
                trace_id=trace_id,
                log_group=log_group,
                log_stream=log_stream,
                event_bus=event_bus,
                event_source=event_source,
                metrics_namespace=metrics_namespace,
                log_putting=log_putting,
                metric_putting=metric_putting,
                event_putting=event_putting,
            )
            root_scope = scope

        else:
            assert scope.parent_id in scopes  # nosec: B101
            scope_store = scopes[scope.parent_id].child(scope)

        scopes[scope.scope_id] = scope_store

        scope_store.record_scope_start()

        return str(scope_store.identifier.scope_id)

    def scope_exiting(
        scope: ContextIdentifier,
        /,
        *,
        exception: BaseException | None,
    ) -> None:
        nonlocal root_scope
        nonlocal trace_id
        assert root_scope is not None  # nosec: B101
        assert scope.scope_id in scopes  # nosec: B101

        scope_store: ScopeStore = scopes[scope.scope_id]
        scope_store.exit()
        scope_store.record_scope_end(exception)

        if not scope_store.try_complete():
            return  # not completed yet or already completed

        # try complete parent scopes
        if scope != root_scope:
            parent_id: UUID = scope.parent_id
            while scopes[parent_id].try_complete():
                if scopes[parent_id].identifier == root_scope:
                    break

                parent_id = scopes[parent_id].identifier.parent_id

        # check for root completion
        if scopes[root_scope.scope_id].completed:
            trace_id = uuid4()
            root_scope = None
            scopes.clear()

    return Observability(
        trace_identifying=trace_identifying,
        log_recording=log_recording,
        metric_recording=metric_recording,
        event_recording=event_recording,
        attributes_recording=attributes_recording,
        scope_entering=scope_entering,
        scope_exiting=scope_exiting,
    )


class ScopeStore:
    __slots__ = (
        "_completion_time",
        "_end_time",
        "_start_time",
        "event_bus",
        "event_putting",
        "event_source",
        "identifier",
        "log_group",
        "log_putting",
        "log_stream",
        "metric_putting",
        "metrics_namespace",
        "nested",
        "trace_id",
    )

    def __init__(
        self,
        identifier: ContextIdentifier,
        /,
        *,
        trace_id: UUID,
        log_group: str,
        log_stream: str,
        event_bus: str,
        event_source: str,
        metrics_namespace: str,
        log_putting: AWSCloudwatchLogPutting,
        metric_putting: AWSCloudwatchMetricPutting,
        event_putting: AWSCloudwatchEventPutting,
    ) -> None:
        self.identifier: ContextIdentifier = identifier
        self.trace_id: UUID = trace_id
        self.log_group: str = log_group
        self.log_stream: str = log_stream
        self.event_bus: str = event_bus
        self.event_source: str = event_source
        self.metrics_namespace: str = metrics_namespace
        self.log_putting: AWSCloudwatchLogPutting = log_putting
        self.metric_putting: AWSCloudwatchMetricPutting = metric_putting
        self.event_putting: AWSCloudwatchEventPutting = event_putting
        self.nested: MutableSequence[ScopeStore] = []
        self._start_time: float = monotonic()
        self._end_time: float | None = None
        self._completion_time: float | None = None

    def child(
        self,
        identifier: ContextIdentifier,
        /,
    ) -> Self:
        child: Self = self.__class__(
            identifier,
            trace_id=self.trace_id,
            log_group=self.log_group,
            log_stream=self.log_stream,
            event_bus=self.event_bus,
            event_source=self.event_source,
            metrics_namespace=self.metrics_namespace,
            log_putting=self.log_putting,
            metric_putting=self.metric_putting,
            event_putting=self.event_putting,
        )
        self.nested.append(child)
        return child

    @property
    def exited(self) -> bool:
        return self._end_time is not None

    def exit(self) -> None:
        assert self._end_time is None  # nosec: B101
        self._end_time = monotonic()

    @property
    def completed(self) -> bool:
        return self._completion_time is not None and all(nested.completed for nested in self.nested)

    def try_complete(self) -> bool:
        if self._end_time is None:
            return False  # not elegible for completion yet

        if self._completion_time is not None:
            return False  # already completed

        if not all(nested.completed for nested in self.nested):
            return False  # nested not completed

        self._completion_time = monotonic()
        return True  # successfully completed

    def record_log(
        self,
        message: str,
        /,
        *,
        level: ObservabilityLevel,
        exception: BaseException | None,
    ) -> None:
        attributes: MutableMapping[str, Any] = {
            "scope.name": self.identifier.name,
            "scope.id": self.identifier.scope_id.hex,
        }
        parent_span_id: str | None = _parent_span_id(self.identifier)
        if parent_span_id is not None:
            attributes["span.parent_id"] = parent_span_id

        if exception is not None:
            attributes["exception.type"] = exception.__class__.__name__
            attributes["exception.message"] = str(exception)

        ctx.spawn_background(
            self.log_putting,
            log_stream=self.log_stream,
            log_group=self.log_group,
            message=_json_dumps(
                {
                    "time_unix_nano": time_ns(),
                    "severity_text": level.name,
                    "severity_number": level.value,
                    "body": message,
                    "trace_id": self.trace_id.hex,
                    "span_id": _span_id(self.identifier),
                    "attributes": _sanitized_attributes(attributes),
                }
            ),
        )

    def record_exception(
        self,
        exception: BaseException,
        /,
    ) -> None:
        ctx.spawn_background(
            self.event_putting,
            event_bus=self.event_bus,
            event_source=self.event_source,
            detail_type="exception",
            detail=_json_dumps(
                {
                    "time_unix_nano": time_ns(),
                    "trace_id": self.trace_id.hex,
                    "span_id": _span_id(self.identifier),
                    "name": "exception",
                    "attributes": _sanitized_attributes(
                        {
                            "scope.name": self.identifier.name,
                            "scope.id": self.identifier.scope_id.hex,
                            "exception.type": exception.__class__.__name__,
                            "exception.message": str(exception),
                            "exception.stacktrace": "".join(
                                traceback.format_exception(
                                    exception.__class__,
                                    exception,
                                    exception.__traceback__,
                                )
                            ),
                        }
                    ),
                }
            ),
        )

    def record_event(
        self,
        event: str,
        /,
        *,
        level: ObservabilityLevel,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        ctx.spawn_background(
            self.event_putting,
            event_bus=self.event_bus,
            event_source=self.event_source,
            detail_type="event",
            detail=_json_dumps(
                {
                    "time_unix_nano": time_ns(),
                    "trace_id": self.trace_id.hex,
                    "span_id": _span_id(self.identifier),
                    "severity_text": level.name,
                    "severity_number": level.value,
                    "name": event,
                    "attributes": _sanitized_attributes(
                        {
                            "scope.name": self.identifier.name,
                            "scope.id": self.identifier.scope_id.hex,
                            **attributes,
                        }
                    ),
                }
            ),
        )

    def record_metric(
        self,
        name: str,
        /,
        *,
        value: float | int,
        unit: str | None,
        kind: ObservabilityMetricKind,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        metric_attributes: dict[str, ObservabilityAttribute] = dict(
            _sanitized_attributes(attributes)
        )
        metric_attributes.setdefault("otel.metric.kind", kind)
        ctx.spawn_background(
            self.metric_putting,
            namespace=self.metrics_namespace,
            metric=name,
            value=value,
            unit=unit,
            attributes=metric_attributes,
        )

    def record_attributes(
        self,
        attributes: Mapping[str, ObservabilityAttribute],
        /,
        *,
        level: ObservabilityLevel,
    ) -> None:
        if not attributes:
            return

        self.record_event(
            "attributes",
            level=level,
            attributes=attributes,
        )

    def record_scope_start(self) -> None:
        self.record_event(
            "span.start",
            level=ObservabilityLevel.INFO,
            attributes={"span.name": self.identifier.name},
        )

    def record_scope_end(
        self,
        exception: BaseException | None,
        /,
    ) -> None:
        if exception is not None:
            self.record_exception(exception)

        self.record_event(
            "span.end",
            level=ObservabilityLevel.INFO,
            attributes={
                "span.name": self.identifier.name,
                "status.code": "ERROR" if exception is not None else "OK",
            },
        )


def _span_id(
    identifier: ContextIdentifier,
    /,
) -> str:
    return identifier.scope_id.hex[-16:]


def _parent_span_id(
    identifier: ContextIdentifier,
    /,
) -> str | None:
    if identifier.is_root:
        return None

    return identifier.parent_id.hex[-16:]


def _sanitized_attributes(
    attributes: Mapping[str, Any],
    /,
) -> Mapping[str, Any]:
    sanitized: dict[str, Any] = {}
    for key, value in attributes.items():
        if value is None or value is MISSING:
            continue  # skip missing/empty

        if isinstance(value, Sequence) and not isinstance(value, str | bytes | bytearray):
            elements: list[Any] = []
            for item in cast(Sequence[Any], value):
                if item is None or item is MISSING:
                    continue  # skip missing/empty

                assert isinstance(item, str | float | int | bool)  # nosec: B101
                elements.append(item)

            if not elements:
                continue  # skip missing/empty

            sanitized[key] = tuple(elements)

        elif isinstance(value, Mapping):
            for name, item in cast(Mapping[str, Any], value).items():
                if item is None or item is MISSING:
                    continue  # skip missing/empty

                assert isinstance(item, str | float | int | bool)  # nosec: B101
                sanitized[f"{key}.{name}"] = item

        else:
            assert isinstance(value, str | float | int | bool)  # nosec: B101
            sanitized[key] = value

    return sanitized


def _json_dumps(
    payload: Mapping[str, Any],
    /,
) -> str:
    return json.dumps(
        payload,
        ensure_ascii=True,
        separators=(",", ":"),
        default=str,
    )



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.utils import FORMAT_INSTRUCTION, extract_evaluation_result
from draive.multimodal import Multimodal, MultimodalContent
from draive.stages import Stage

__all__ = ("coherence_evaluator",)


INSTRUCTION: str = f"""\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate the EVALUATED content using solely a coherence metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is coherence - a collective quality of the content.
We align this dimension with the DUC (Document Understanding Conference) quality question of structure and coherence, whereby the content should be well-structured and well-organized.
EVALUATED content should not just be a heap of related information, but should build from part to part into a coherent body of information about the topic.
</EVALUATION_CRITERIA>
{{guidelines}}
<RATING>
Assign a coherence score using exact name of one of the following values:
- "poor" is very low coherence, the content is chaotic, lacking logical connections between parts.
- "fair" is low coherence, some connections are visible, but the overall structure is weak.
- "good" is moderate coherence, the content has a noticeable structure, but with some shortcomings.
- "excellent" is high coherence, the content is well-organized with minor imperfections.
- "perfect" is very high coherence, the content is exemplarily structured, with smooth transitions between ideas.
Use the "none" value for content that cannot be rated at all.
</RATING>

{FORMAT_INSTRUCTION}
"""  # noqa: E501


@evaluator(name="coherence")
async def coherence_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Input was empty!"},
        )

    if not reference:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Reference was empty!"},
        )

    return extract_evaluation_result(
        await Stage.completion(
            MultimodalContent.of(
                "<REFERENCE>",
                reference,
                "</REFERENCE>\n<EVALUATED>",
                evaluated,
                "</EVALUATED>",
            ),
            instructions=INSTRUCTION.format(
                guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n" if guidelines else "",
            ),
        ).execute()
    )



from collections.abc import Mapping, Sequence
from typing import Any, Protocol, Self, final, runtime_checkable

from haiway import META_EMPTY, Meta, MetaValues, State

from draive.multimodal.content import Multimodal

__all__ = (
    "Template",
    "TemplateDeclaration",
    "TemplateDefining",
    "TemplateListing",
    "TemplateLoading",
    "TemplateMissing",
)


class TemplateMissing(Exception):
    """Raised when a requested template declaration or content cannot be found.

    Parameters
    ----------
    identifier
        Identifier of the missing template.
    """

    __slots__ = ("identifier",)

    def __init__(
        self,
        *,
        identifier: str,
    ) -> None:
        super().__init__(f"Missing template - {identifier}")
        self.identifier: str = identifier


@final
class Template(State):
    """Immutable template reference describing how to render multimodal content.

    Templates reference concrete content by identifier and optionally carry default
    argument bindings that will be merged with arguments provided during resolution.

    Parameters
    ----------
    identifier
        Unique name of the template to load from a repository.
    arguments
        Default argument mapping applied on top of call-time arguments.
    meta
        Supplemental metadata propagated to the underlying repository implementation.
    """

    @classmethod
    def of(
        cls,
        identifier: str,
        /,
        *,
        arguments: Mapping[str, Multimodal] | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """Create a template reference with optional defaults.

        Parameters
        ----------
        identifier
            Unique name of the template.
        arguments
            Default argument mapping applied when resolving the template.
        meta
            Optional metadata forwarded to repository calls.

        Returns
        -------
        Self
            New template reference instance.
        """
        return cls(
            identifier=identifier,
            arguments=arguments if arguments is not None else {},
            meta=Meta.of(meta),
        )

    identifier: str
    arguments: Mapping[str, Multimodal]
    meta: Meta = META_EMPTY

    def with_arguments(
        self,
        **arguments: Multimodal,
    ) -> Self:
        """Return a new template augmented with additional default arguments.

        Parameters
        ----------
        **arguments
            Additional argument bindings to merge with existing defaults.

        Returns
        -------
        Self
            Template instance containing merged argument defaults.
        """
        if not arguments:
            return self

        return self.__class__(
            identifier=self.identifier,
            arguments={**self.arguments, **arguments},
            meta=self.meta,
        )

    def with_meta(
        self,
        meta: Meta | MetaValues,
        /,
    ) -> Self:
        """Return a new template with metadata merged into the current meta.

        Parameters
        ----------
        meta
            Extra metadata values to merge with the existing metadata.

        Returns
        -------
        Self
            Template instance carrying the merged metadata.
        """
        if not meta:
            return self

        return self.__class__(
            identifier=self.identifier,
            arguments=self.arguments,
            meta=self.meta.merged_with(meta),
        )


@final
class TemplateDeclaration(State):
    """Immutable template declaration describing user-facing template metadata.

    Declarations surface template descriptions and declared variables without
    requiring the template content itself to be loaded.

    Parameters
    ----------
    identifier
        Unique name of the template declaration.
    description
        Optional human-readable summary of the template purpose.
    variables
        Mapping of variable names to description strings.
    meta
        Supplemental metadata preserved when invoking repository operations.
    """

    @classmethod
    def of(
        cls,
        identifier: str,
        /,
        *,
        description: str | None = None,
        variables: Mapping[str, str] | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """Create a template declaration with optional description and variables.

        Parameters
        ----------
        identifier
            Unique name of the template declaration.
        description
            Optional human-readable description.
        variables
            Mapping of variable names to their descriptions.
        meta
            Optional metadata forwarded to repository calls.

        Returns
        -------
        Self
            New template declaration instance.
        """
        return cls(
            identifier=identifier,
            description=description,
            variables=variables if variables is not None else {},
            meta=Meta.of(meta),
        )

    identifier: str
    description: str | None = None
    variables: Mapping[str, str]
    meta: Meta = META_EMPTY


@runtime_checkable
class TemplateListing(Protocol):
    """Callable protocol returning available template declarations."""

    async def __call__(
        self,
        **extra: Any,
    ) -> Sequence[TemplateDeclaration]: ...


@runtime_checkable
class TemplateLoading(Protocol):
    """Callable protocol loading template content for the given identifier."""

    async def __call__(
        self,
        identifier: str,
        meta: Meta,
        **extra: Any,
    ) -> str | None: ...


@runtime_checkable
class TemplateDefining(Protocol):
    """Callable protocol that upserts template content and metadata."""

    async def __call__(
        self,
        identifier: str,
        description: str | None,
        content: str,
        variables: Mapping[str, str],
        meta: Meta,
        **extra: Any,
    ) -> None: ...



from collections.abc import Mapping, Sequence
from datetime import UTC, datetime
from time import time_ns
from typing import Any

from botocore.exceptions import ClientError  # pyright: ignore[reportMissingModuleSource]
from haiway import MISSING, ObservabilityAttribute, asynchronous

from draive.aws.api import AWSAPI
from draive.aws.types import (
    AWSAccessDenied,
    AWSError,
    AWSResourceNotFound,
)

__all__ = ("AWSCloudwatchMixin",)


class AWSCloudwatchMixin(AWSAPI):
    def __init__(
        self,
        region_name: str | None = None,
        access_key_id: str | None = None,
        secret_access_key: str | None = None,
        profile_name: str | None = None,
    ) -> None:
        super().__init__(
            region_name=region_name,
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            profile_name=profile_name,
        )

    async def put_log(
        self,
        *,
        log_group: str,
        log_stream: str,
        message: str,
    ) -> None:
        await self._put_log(
            log_group=log_group,
            log_stream=log_stream,
            message=message,
        )

    @asynchronous
    def _put_log(
        self,
        *,
        log_group: str,
        log_stream: str,
        message: str,
    ) -> None:
        try:
            self._cloudwatch_logs_client.put_log_events(
                logGroupName=log_group,
                logStreamName=log_stream,
                logEvents=[
                    {
                        "timestamp": time_ns() // 1_000_000,  # epoch ms (UTC)
                        "message": message,
                    }
                ],
            )

        except ClientError as exc:
            raise translate_cloudwatch_error(
                error=exc,
                service="logs",
                operation="put_log",
                resource=f"{log_group}/{log_stream}",
            ) from exc

    async def put_event(
        self,
        *,
        event_bus: str,
        event_source: str,
        detail_type: str,
        detail: str,
    ) -> None:
        await self._put_event(
            event_bus=event_bus,
            event_source=event_source,
            detail_type=detail_type,
            detail=detail,
        )

    @asynchronous
    def _put_event(
        self,
        *,
        event_bus: str,
        event_source: str,
        detail_type: str,
        detail: str,
    ) -> None:
        try:
            self._eventbridge_client.put_events(
                Entries=[
                    {
                        "EventBusName": event_bus,
                        "Source": event_source,
                        "DetailType": detail_type,
                        "Detail": detail,
                        "Time": datetime.now(tz=UTC),
                    }
                ],
            )

        except ClientError as exc:
            raise translate_cloudwatch_error(
                error=exc,
                service="events",
                operation="put_events",
                resource=event_bus,
            ) from exc

    async def put_metric(
        self,
        *,
        namespace: str,
        metric: str,
        value: float | int,
        unit: str | None,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        await self._put_metric(
            namespace=namespace,
            metric=metric,
            value=value,
            unit=unit,
            attributes=attributes,
        )

    @asynchronous
    def _put_metric(
        self,
        *,
        namespace: str,
        metric: str,
        value: float | int,
        unit: str | None,
        attributes: Mapping[str, ObservabilityAttribute],
    ) -> None:
        try:
            dimensions = format_metric_dimensions(attributes)
            metric_data: dict[str, Any] = {
                "MetricName": metric,
                "Value": value,
            }

            if dimensions:
                metric_data["Dimensions"] = dimensions

            if unit:
                metric_data["Unit"] = unit

            self._cloudwatch_client.put_metric_data(
                Namespace=namespace,
                MetricData=[metric_data],
            )

        except ClientError as exc:
            raise translate_cloudwatch_error(
                error=exc,
                service="cloudwatch",
                operation="put_metric_data",
                resource=namespace,
            ) from exc


def format_metric_dimensions(
    attributes: Mapping[str, ObservabilityAttribute],
) -> list[dict[str, str]]:
    dimensions: list[dict[str, str]] = []
    for key, value in _sanitize_attributes(attributes).items():
        rendered: str | None = _render_dimension_value(value)
        if rendered is None:
            continue

        dimensions.append(
            {
                "Name": _truncate_dimension(key),
                "Value": _truncate_dimension(rendered),
            }
        )

    return dimensions


def _render_dimension_value(
    value: ObservabilityAttribute,
) -> str | None:
    if value is None or value is MISSING:
        return None

    if isinstance(value, str):
        return value

    elif isinstance(value, Sequence):
        items: list[str] = []
        for item in value:
            items.append(str(item))

        if not items:
            return None

        return ",".join(items)

    return str(value)


def _sanitize_attributes(
    attributes: Mapping[str, ObservabilityAttribute],
) -> Mapping[str, ObservabilityAttribute]:
    sanitized: dict[str, ObservabilityAttribute] = {}
    for key, value in attributes.items():
        if value is None or value is MISSING:
            continue

        if isinstance(value, str):
            sanitized[key] = value

        elif isinstance(value, Sequence):
            if not value:
                continue

            sanitized[key] = value

        else:
            sanitized[key] = value

    return sanitized


def _truncate_dimension(
    value: str,
    *,
    max_length: int = 255,
) -> str:
    if len(value) <= max_length:
        return value

    return value[:max_length]


def translate_cloudwatch_error(
    *,
    error: ClientError,
    service: str,
    operation: str,
    resource: str,
) -> Exception:
    error_info: Mapping[str, Any] = getattr(error, "response", {}).get("Error", {})
    response_metadata: Mapping[str, Any] = getattr(error, "response", {}).get(
        "ResponseMetadata", {}
    )

    code: str = str(error_info.get("Code") or "").strip()
    message: str = str(error_info.get("Message") or str(error)).strip()
    status_code: int | None = response_metadata.get("HTTPStatusCode")

    normalized_code: str = code.lower()
    uri: str = f"{service}://{resource}"
    contextual_message: str = (
        f"{message} [provider=aws_{service} operation={operation} resource={resource}]"
    )

    if normalized_code in {
        "accessdenied",
        "accessdeniedexception",
        "invalidaccesskeyid",
        "signaturedoesnotmatch",
    } or status_code in {401, 403}:
        return AWSAccessDenied(
            uri=f"{service}://{resource}",
            code=code or None,
            message=contextual_message,
        )

    if (
        status_code == 404 or normalized_code in {"resourcenotfoundexception", "notfound"}  # noqa: PLR2004
    ):
        return AWSResourceNotFound(
            uri=uri,
            code=code or None,
            message=contextual_message,
        )

    return AWSError(
        uri=uri,
        code=code or None,
        message=contextual_message,
    )



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.utils import FORMAT_INSTRUCTION, extract_evaluation_result
from draive.multimodal import Multimodal, MultimodalContent
from draive.stages import Stage

__all__ = ("consistency_evaluator",)


INSTRUCTION: str = f"""\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate the EVALUATED content using solely a consistency metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is consistency - a factual alignment between the REFERENCE and the EVALUATED content.
A factually consistent content contains only elements that are entailed by the REFERENCE content.
</EVALUATION_CRITERIA>
{{guidelines}}
<RATING>
Assign a consistency score using exact name of one of the following values:
- "poor" is very low consistency, the content contains multiple hallucinated facts or significant misalignments with the reference content.
- "fair" is low consistency, the content has several instances of information not supported by the reference content.
- "good" is moderate consistency, the content is mostly consistent but contains a few unsupported statements.
- "excellent" is high consistency, the content is largely consistent with minor discrepancies.
- "perfect" is very high consistency, the content is fully consistent with the reference content, containing only supported information.
Use the "none" value for content that cannot be rated at all.
</RATING>

{FORMAT_INSTRUCTION}
"""  # noqa: E501


@evaluator(name="consistency")
async def consistency_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Input was empty!"},
        )

    if not reference:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Reference was empty!"},
        )

    return extract_evaluation_result(
        await Stage.completion(
            MultimodalContent.of(
                "<REFERENCE>",
                reference,
                "</REFERENCE>\n<EVALUATED>",
                evaluated,
                "</EVALUATED>",
            ),
            instructions=INSTRUCTION.format(
                guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n" if guidelines else "",
            ),
        ).execute()
    )



from typing import Any, overload

from haiway import State, statemethod

from draive.guardrails.safety.regex import guardrails_regex_sanitizer
from draive.guardrails.safety.types import (
    GuardrailsSafetyException,
    GuardrailsSafetySanitization,
)
from draive.guardrails.types import GuardrailsException, GuardrailsFailure
from draive.multimodal import Multimodal, MultimodalContent

__all__ = ("GuardrailsSafety",)


class GuardrailsSafety(State):
    @overload
    @classmethod
    async def sanitize(
        cls,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> MultimodalContent: ...

    @overload
    async def sanitize(
        self,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> MultimodalContent: ...

    @statemethod
    async def sanitize(
        self,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> MultimodalContent:
        content = MultimodalContent.of(content)
        try:
            return await self.sanitization(
                content,
                **extra,
            )

        except GuardrailsSafetyException:
            raise

        except GuardrailsException as exc:
            raise GuardrailsSafetyException(
                f"Safety guardrails triggered: {exc}",
                content=content,
                reason=str(exc),
                meta=exc.meta,
            ) from exc

        except Exception as exc:
            raise GuardrailsFailure(
                f"Safety guardrails failed: {exc}",
                cause=exc,
                meta={"error_type": exc.__class__.__name__},
            ) from exc

    # use default set of regex rules for sanitization
    sanitization: GuardrailsSafetySanitization = guardrails_regex_sanitizer()
